# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
This module implements the generic class for loading/dumping a dataset from/to
file.
"""
import numpy as np
import numbers
import hashlib

from .svmlight_format import load_svmlight_file, dump_svmlight_file


class Dataset(object):
    """
    This class describe the dataset object, with its utility and features

    Attributes
    ----------
    X : numpy 2d array of float
        It is a dense numpy matrix of shape (n_instances, n_features),
    y : numpy 1d array of float
        It is a ndarray of shape (n_instances,) with the gold label
    query_ids : numpy 1d array of int
        It is a ndarray of shape(n_queries,)
    query_offsets : numpy 1d array of int
        It is a ndarray of shape(n_queries+1, ) with the start and end offsets
        of each query. In particular. the i-th query has indices ranging in
        [ query_offsets[i], query_offsets[i+1] ), with the latter excluded.
    name : str
        The name to give to the dataset
    n_instances : int
        The number of instances in the dataset
    n_features : int
        The number of features in the dataset
    n_queries : int
        The number of queries in the dataset
    """

    def __init__(self, X, y, query_ids, name=None):
        """
        This module implements the generic class for loading/dumping a dataset
        from/to file.

        Parameters
        ----------
        X : numpy.ndarray
            The matrix with feature values
        y : numpy.array
            The vector with label values
        query_ids : numpy.array
            The vector with the query_id for each sample.
        """
        if query_ids.size != X.shape[0]:
            raise Exception("query_ids has wrong size. Expected %s but got %s" % (X.shape[0], query_ids.size))

        # convert from query_ids per sample to query offset
        self.query_ids, self.query_offsets = \
            np.unique(query_ids, return_index=True)

        # resort the arrays per offset (if the file does not contains qids in
        # order, the np.unique will return qids with a different ordering...
        idx_sort = np.argsort(self.query_offsets)
        self.query_ids = self.query_ids[idx_sort]
        self.query_offsets = self.query_offsets[idx_sort]

        self.query_offsets = np.append(self.query_offsets, query_ids.size)

        self.X, self.y = X, y
        self.name = "Dataset %s" % (self.X.shape,)
        if name is not None:
            self.name = name

        self.n_instances = self.y.size
        self.n_features = self.X.shape[1]
        self.n_queries = self.query_ids.size

    @staticmethod
    def load(f, name=None, format="svmlight"):
        """
        This static method implements the loading of a dataset from file.

        Parameters
        ----------
        f : {str, file-like, int}
            (Path to) a file to load. If a path ends in ".gz" or ".bz2", it will
            be uncompressed on the fly. If an integer is passed, it is assumed
            to be a file descriptor. A file-like or file descriptor will not be
            closed by this function. A file-like object must be opened in
            binary mode.
        name : str
            The name to be given to the current dataset
        format : str
            The format of the dataset file to load (actually supported is only
            "svmlight" format)

        Returns
        -------
        dataset : Dataset
            The dataset read from file
        """
        if format == "svmlight":
            X, y, query_ids = load_svmlight_file(f, query_id=True)
        else:
            raise TypeError("Dataset format %s is not yet supported!" % format)
        return Dataset(X, y, query_ids, name)

    def subset_features(self, features):
        """
        Create a new Dataset with only the features identified by the given
        features parameters (indices). It is useful for performing feature
        selection.

        Parameters
        ----------
        features : numpy array or list
            The indices of the features to select in the resulting dataset

        Returns
        -------
        dataset : rankeval.dataset.Dataset
            The resulting dataset with the given subset of features
        """
        return Dataset(self.X[:, features].copy(),
                       self.y,
                       self.get_qids_dataset(),
                       name=self.name)

    def dump(self, f, format="svmlight"):
        """
        This method implements the writing of a previously loaded dataset
        according to the given format on file

        Parameters
        ----------
        f : {str, file-like, int}
            (Path to) a file to dump. If a path ends in ".gz" or ".bz2", it will
            be compressed on the fly. If an integer is passed, it is assumed
            to be a file descriptor. A file-like or file descriptor will not be
            closed by this function. A file-like object must be opened in
            text mode.
        format : str
            The format to use for dumping the dataset on file (actually
            supported is only "svmlight" format)
        """
        # we need to unroll the query_ids and query_offsets.
        # They are represented compact: they report only the query ids and the
        # offsets where each query starts and ends.
        query_ids = np.ndarray(self.n_instances, dtype=np.int32)
        for qid, start_offset, end_offset in self.query_iterator():
            for idx in np.arange(start_offset, end_offset):
                query_ids[idx] = qid

        if format == "svmlight":
            dump_svmlight_file(self.X, self.y, f, query_ids)
        else:
            raise TypeError("Dataset format %s is not yet supported!" % format)

    def split(self, train_size, vali_size=0, random_state=None):
        """
        This method splits the dataset into train/validation/test partition.
        It shuffle the query ids before partitioning. If vali_size=0, it means
        the method will not create a validation set, thus returning only
        train and test sets. Otherwise it will return train/vali/test sets.

        Parameters
        ----------
        train_size : float
            The ratio of query ids in the training set. It should be between
            0 and 1.
        vali_size : float
            The ratio of query ids in the validation set. It should be between
            0 and 1. 0 means no validation to be created.
        random_state : int
            If int, random_state is the seed used by the random number
            generator. If RandomState instance, random_state is the random
            number generator. If None, the random number generator is the
            RandomState instance used by np.random.

        Returns
        -------
        (train, vali, test) datasets : tuple of rankeval.dataset.Dataset
            The resulting datasets with the given fraction of query ids in each
            partition.
        """

        if train_size < 0 or train_size > 1 or (train_size + vali_size) > 1:
            raise Exception("train and/or validation sizes are not correct!")

        train_qn = int(round(train_size * self.n_queries))
        vali_qn = int(round(vali_size * self.n_queries))
        test_qn = self.n_queries - train_qn - vali_qn

        qid_map = np.ndarray(self.n_instances, dtype=np.uint32)
        for qid, start_offset, end_offset in self.query_iterator():
            for idx in np.arange(start_offset, end_offset):
                qid_map[idx] = qid

        # add queries shuffling
        rng = Dataset._check_random_state(random_state)
        qids_permutation = rng.permutation(self.query_ids)

        train_qid = qids_permutation[:train_qn]
        vali_qid = qids_permutation[train_qn:train_qn+vali_qn]
        test_qid = qids_permutation[-test_qn:]

        train_mask = np.in1d(qid_map, train_qid)
        vali_mask = np.in1d(qid_map, vali_qid)
        test_mask = np.in1d(qid_map, test_qid)

        train_dataset = Dataset(self.X[train_mask], self.y[train_mask],
                                qid_map[train_mask], name=self.name + ' Train')
        if vali_size:
            vali_dataset = Dataset(self.X[vali_mask], self.y[vali_mask],
                                   qid_map[vali_mask], name=self.name + ' Vali')
        test_dataset = Dataset(self.X[test_mask], self.y[test_mask],
                               qid_map[test_mask], name=self.name + ' Test')

        if not vali_size:
            return train_dataset, test_dataset
        else:
            return train_dataset, vali_dataset, test_dataset

    def subset(self, query_ids, name=None):
        """
        This method return a subset of the dataset according to the query_ids
        parameter.

        Parameters
        ----------
        query_ids : numpy 1d array of int
            It is a ndarray with the query_ids to select
        name : str
            The name to give to the dataset

        Returns
        -------
        datasets : rankeval.dataset.Dataset
            The resulting dataset with only the query_ids requested
        """
        qid_map = self.get_qids_dataset()
        mask = np.in1d(qid_map, query_ids)

        return Dataset(self.X[mask], self.y[mask],
                       qid_map[mask], name=name)

    def query_iterator(self):
        """
        This method implements and iterator over the offsets of the query_ids
        in the dataset.

        Returns
        -------
        offsets : tuple of (int, int, int)
            The query_id and the row index of instances belonging to the query.
            The two indices represent (start, end) offsets.

        """
        for i in np.arange(self.n_queries):
            yield self.query_ids[i], \
                  self.query_offsets[i], self.query_offsets[i+1]

    def get_query_sizes(self):
        """
        This method return the size of each query set.

        Returns
        -------
        sizes : numpy 1d array of int
            It is a ndarray of shape (n_queries,)
        """
        return np.ediff1d(self.query_offsets)

    def get_qids_dataset(self, dtype=np.int32):
        """
        This method returns the query ids array in linear representation, i.e.,
        with the qid of each instance. Useful for creating a new dataset
        starting from a different one.

        Returns
        -------
        query_ids : numpy 1d array
            It is a ndarray of shape (n_instances,)
        """
        query_ids = np.empty(shape=self.n_instances, dtype=dtype)
        for qid, start_offset, end_offset in self.query_iterator():
            query_ids[start_offset:end_offset] = qid
        return query_ids

    @staticmethod
    def _check_random_state(seed):
        """
        Turn seed into a np.random.RandomState instance (took for sklearn)

        Parameters
        ----------
        seed : None | int | instance of RandomState
            If seed is None, return the RandomState singleton used by np.random.
            If seed is an int, return a new RandomState instance seeded with it.
            If seed is already a RandomState instance, return it.
            Otherwise raise ValueError.
        """
        if seed is None or seed is np.random:
            return np.random.mtrand._rand
        if isinstance(seed, (numbers.Integral, np.integer)):
            return np.random.RandomState(seed)
        if isinstance(seed, np.random.RandomState):
            return seed
        raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                         ' instance' % seed)

    def __str__(self):
        return self.name

    def __hash__(self):

        h = hashlib.md5()

        for arr in [self.X, self.y, self.query_ids]:
            h.update(arr)

        return int(h.hexdigest(), 16)

    def __eq__(self, other):
        # use != instead of == because it is more efficient for sparse matrices:
        x_eq = not(self.X != other.X).any()
        return x_eq and (self.y == other.y).all() and \
               (self.query_ids == other.query_ids).all()

    def __ne__(self, other):
        # Not strictly necessary, but to avoid having both x==y and x!=y
        # True at the same time
        return not(self == other)



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Class providing the implementation for loading/storing a Scikit-Learn model
from/to file. The model has to be saved using the export_scikit_model method:
.. code-block:: python
    from sklearn import ensemble
    from rankeval.model import ProxyScikitLearn
    ...
    gbrt = ensemble.GradientBoostingRegressor(...)
    gbrt.fit(X, y)
    ProxyScikitLearn.export_scikit_model(model=gbrt, file_path=model_path)
    gbrt_rankeval = RTEnsemble(file_path=model_path, format='ScikitLearn')

The Scikit-Learn project is described here:
    http://scikit-learn.org/
"""

import re

import numpy as np

from .rt_ensemble import RTEnsemble

base_score_reg = re.compile("^base_score=(.+)$")
learning_rate_reg = re.compile("^learning_rate=(.+)$")
tree_reg = re.compile("^booster\[(\d+)\]")
node_reg = re.compile("(\d+):\[f(\d+)<=(.*)\]")
leaf_reg = re.compile("(\d+):leaf=(.+)$")


class ProxyScikitLearn(object):
    """
    Class providing the implementation for loading/storing a Scikit-Learn model
    from/to file.
    """

    @staticmethod
    def load(file_path, model):
        """
        Load the model from the file identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has been saved
        model : RTEnsemble
            The model instance to fill
        """
        n_trees, n_nodes = ProxyScikitLearn._count_nodes(file_path)
        # Initialize the model and allocate the needed space
        # given the shape and size of the ensemble
        model.initialize(n_trees, n_nodes)

        root_node = 0
        num_nodes = 0
        learning_rate = 1
        queue = list()
        with open(file_path, 'r') as f:
            for line in f:

                match = base_score_reg.match(line)
                if match:
                    model.base_score = float(match.group(1))
                    continue

                match = learning_rate_reg.match(line)
                if match:
                    model.learning_rate = float(match.group(1))
                    continue

                match_tree = tree_reg.match(line)
                if match_tree:
                    assert(len(queue) == 0)
                    curr_tree = int(match_tree.group(1))
                    root_node += num_nodes
                    num_nodes = 0
                    model.trees_root[curr_tree] = root_node
                    model.trees_weight[curr_tree] = 1
                    continue

                match_node = node_reg.search(line)
                if match_node:
                    node_id = int(match_node.group(1).strip()) + root_node
                    feature_id = int(match_node.group(2).strip())
                    threshold = float(match_node.group(3).strip())

                    model.trees_nodes_feature[node_id] = feature_id
                    model.trees_nodes_value[node_id] = threshold

                match_leaf = leaf_reg.search(line)
                if match_leaf:
                    node_id = int(match_leaf.group(1).strip()) + root_node
                    leaf_value = float(match_leaf.group(2).strip())
                    model.trees_nodes_value[node_id] = leaf_value

                if match_node or match_leaf:
                    num_nodes += 1
                    if len(queue) > 0:
                        parent_id, child = queue.pop()
                        if child == 'L':
                            model.trees_left_child[parent_id] = node_id
                        else:
                            model.trees_right_child[parent_id] = node_id

                if match_node:
                    # two elements in the queue for the left and right children
                    # Each element is identified by a node_id and the indication
                    # of being the left or right child.
                    queue.extend([(node_id, 'R'), (node_id, 'L')])

    @staticmethod
    def save(file_path, model):
        """
        Save the model onto the file identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has to be saved
        model : RTEnsemble
            The model RTEnsemble model to save on file

        Returns
        -------
        status : bool
            Returns true if the save is successful, false otherwise
        """
        raise NotImplementedError("Feature not implemented!")

    @staticmethod
    def _count_nodes(file_path):
        """
        Count the total number of nodes (both split and leaf nodes)
        in the model identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has been saved

        Returns
        -------
        tuple(n_trees, n_nodes) : tuple(int, int)
            The total number of trees and nodes (both split and leaf nodes)
            in the model identified by file_path.
        """

        n_nodes = 0
        n_trees = 0

        with open(file_path, 'r') as f:
            for line in f:

                match = tree_reg.match(line)
                if match:
                    n_trees += 1
                    continue

                match_node = node_reg.search(line)
                if match_node:
                    n_nodes += 1

                match_leaf = leaf_reg.search(line)
                if match_leaf:
                    n_nodes += 1

        return n_trees, n_nodes

    @staticmethod
    def export_scikit_model(model, file_path):
        if not hasattr(model, 'estimators_'):
            raise TypeError("Only ensemble-based models are supported!")

        if not hasattr(model, 'init_'):
            raise TypeError("Base score missing!")
        if hasattr(model.init_, "quantile"):
            base_score = model.init_.quantile
        elif hasattr(model.init_, "mean"):
            base_score = model.init_.mean
        else:
            raise TypeError("Base score unknown!")


        with open(file_path, 'w') as writer:
            if not hasattr(model, 'init_'):
                raise TypeError("Base score missing!")
            writer.write("base_score=%f\n" % base_score)
            writer.write("learning_rate=%f\n" % model.learning_rate)
            for tree_id, tree in enumerate(model.estimators_.flatten()):
                ProxyScikitLearn._export_tree(writer, tree, tree_id)

    @staticmethod
    def _export_tree(writer, tree, tree_id=0):
        from sklearn.tree import _tree

        if not hasattr(tree, 'tree_'):
            raise TypeError("Only tree-based models are supported!")

        tree_ = tree.tree_
        feature_name = ["f%d" % i for i in np.unique(tree_.feature) if
                        i != _tree.TREE_UNDEFINED]
        writer.write("booster[%d] [%s]:\n" % (tree_id, ' '.join(feature_name)))

        def recurse(node, depth):
            indent = '\t' * depth
            if tree_.feature[node] != _tree.TREE_UNDEFINED:
                name = "f%d" % tree_.feature[node]
                threshold = tree_.threshold[node]
                writer.write("%s%d:[%s<=%f]\n" % (indent, node, name, threshold))
                recurse(tree_.children_left[node], depth + 1)
                # print "%s%d:[%s>%f]" % (indent, node, name, threshold)
                recurse(tree_.children_right[node], depth + 1)
            else:
                if tree_.value[node].size > 1:
                    leaf_value = "c%d" % np.argmax(tree_.value[node])
                else:
                    leaf_value = tree_.value[node].flatten()[0]
                writer.write("%s%d:leaf=%s\n" % (indent, node, leaf_value))

        recurse(0, 1)



import logging
import os
import unittest

from numpy.testing import assert_array_almost_equal, assert_almost_equal

from rankeval.dataset import Dataset
from rankeval.model import RTEnsemble
from rankeval.scoring.scorer import Scorer
from rankeval.test.base import data_dir

model_file = os.path.join(data_dir, "quickrank.model.xml")
data_file = os.path.join(data_dir, "msn1.fold1.test.5k.txt")


class ScoringTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.model = RTEnsemble(model_file, format="QuickRank")
        cls.dataset = Dataset.load(data_file, format="svmlight")
        cls.scorer = Scorer(cls.model, cls.dataset)

    @classmethod
    def tearDownClass(cls):
        del cls.model
        cls.model = None
        del cls.dataset
        cls.dataset = None
        del cls.scorer
        cls.scorer = None

    def test_basic_scoring_values(self):
        self.scorer.score(detailed=False)
        assert_array_almost_equal(self.scorer.get_predicted_scores()[:3],
                                  [0.16549695, 0.07126279, 0.10397919])
        assert_array_almost_equal(self.scorer.get_predicted_scores()[-3:],
                                  [0.13345119, 0.13345119, 0.07126279])

    def test_basic_scoring_sum(self):
        self.scorer.score(detailed=False)
        assert_almost_equal(self.scorer.get_predicted_scores().sum(),
                            598.72852, decimal=5)

    def test_detailed_scoring_values(self):
        self.scorer.score(detailed=True)
        assert_array_almost_equal(
            self.scorer.get_partial_predicted_scores()[:3],
            [[0.06684528,  0.09865167],
            [0.03412888,  0.03713391],
            [0.06684528,  0.03713391]])
        assert_array_almost_equal(
            self.scorer.get_partial_predicted_scores()[-3:],
            [[0.09631728,  0.03713391],
            [0.09631728,  0.03713391],
            [0.03412888,  0.03713391]])

    def test_basic_and_detailed_scoring(self):
        self.scorer.score(detailed=False)
        y_pred_basic = self.scorer.y_pred
        self.scorer.score(detailed=True)
        y_pred_detailed = self.scorer.y_pred
        assert_array_almost_equal(y_pred_basic, y_pred_detailed)

    def test_detailed_scoring_sum(self):
        self.scorer.score(detailed=True)
        assert_almost_equal(self.scorer.get_partial_predicted_scores().sum(),
                            598.72852, decimal=5)
        assert_array_almost_equal(
            self.scorer.get_partial_predicted_scores().sum(axis=0),
            [312.43994141, 286.2948])
        assert_array_almost_equal(
            self.scorer.get_partial_predicted_scores().sum(axis=1)[:3],
            [0.16549695, 0.07126279, 0.10397919])
        assert_array_almost_equal(
            self.scorer.get_partial_predicted_scores().sum(axis=1)[-3:],
            [0.13345119, 0.13345119, 0.07126279])


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s',
                        level=logging.DEBUG)
    unittest.main()



import logging
import os
import unittest

from numpy.testing import assert_equal, assert_array_equal, \
    assert_array_almost_equal

from rankeval.dataset import Dataset
from rankeval.model import ProxyXGBoost
from rankeval.model import RTEnsemble
from rankeval.test.base import data_dir

model_file = os.path.join(data_dir, "ScikitLearn.model.txt")
data_file = os.path.join(data_dir, "msn1.fold1.test.5k.txt")


class ProxyXGBoostTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.model = RTEnsemble(model_file, format="ScikitLearn")
        cls.dataset = Dataset.load(data_file, format="svmlight")

    @classmethod
    def tearDownClass(cls):
        del cls.model
        cls.model = None
        del cls.dataset
        cls.dataset = None

    def test_count_nodes(self):
        n_trees, n_nodes = ProxyXGBoost._count_nodes(model_file)
        # print "Num Trees: %d\nNum Nodes: %d" % (n_trees, n_nodes),
        assert_equal(n_trees, 2)
        assert_equal(n_nodes, 10)
        assert_equal(n_trees, self.model.trees_root.size)
        assert_equal(n_nodes, self.model.trees_nodes_value.size)

    def test_root_nodes(self):
        assert_equal((self.model.trees_root > -1).all(), True,
                     err_msg="Root nodes not set correctly")

    def test_root_nodes_adv(self):
        assert_array_equal(self.model.trees_root, [0, 5],
                           err_msg="Root nodes are not correct")

    def test_split_features(self):
        assert_array_equal(self.model.trees_nodes_feature,
                           [54, -1, 133, -1, -1, 52, 14, -1, -1, -1])

    def test_tree_values(self):
        assert_array_almost_equal(
            self.model.trees_nodes_value,
            [5.769000e-03, -2.264262e-01, 5.000000e-01, 1.872047e-01,
             9.056279e-01, 4.446700e-02, 2.850000e+01, 2.694752e-01,
             4.131481e-01, -2.031448e-01],
            err_msg="Split thresholds or leaf outputs value are not correct")

    def test_left_children(self):
        assert_array_equal(self.model.trees_left_child,
                           [1, -1, 3, -1, -1, 6, 8, -1, -1, -1])

    def test_right_children(self):
        assert_array_equal(self.model.trees_right_child,
                           [2, -1, 4, -1, -1, 7, 9, -1, -1, -1])

    def test_leaf_correctness(self):
        for idx, feature in enumerate(self.model.trees_nodes_feature):
            if feature == -1:
                assert_equal(self.model.trees_left_child[idx], -1,
                             "Left child of a leaf node is not empty (-1)")
                assert_equal(self.model.trees_right_child[idx], -1,
                             "Right child of a leaf node is not empty (-1)")
                assert_equal(self.model.is_leaf_node(idx), True,
                             "Leaf node not detected as a leaf")

    def test_prediction(self):
        y_pred = self.model.score(self.dataset, cache=True)
        assert_array_almost_equal(y_pred[:5],
                                  [0.651668,  0.604406,  0.604406,
                                   0.610305,  0.563043])
        assert_array_almost_equal(y_pred[-5:],
                                  [0.563043,  0.563043,  0.563043,
                                   0.563043,  0.563043])

if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s',
                        level=logging.DEBUG)
    unittest.main()



"""
This package serve the root of the RankEval package.
"""

import os
import io

cur_dir = os.path.dirname(__file__)

__version__ = io.open(os.path.join(cur_dir, 'VERSION'),
                      encoding='utf-8').read().strip()


def is_notebook():
    try:
        from IPython import get_ipython
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except (NameError, ImportError):
        return False      # Probably standard Python interpreter



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Class providing the implementation for loading/storing a CatBoost model
from/to file.

The CatBoost project is described here:
    https://github.com/catboost/catboost

CatBoost allows to save the learned model in several formats (binary, coreml,
etc). Among them, we chose to adopts the Apple CoreML format for reading and
converting a model into the rankeval representation. It s possible to read  the
coreML representation using the coremltools python package. Once read it
provides all the structured information of the ensemble, with split nodes (both
features and thresholds), leaf values and tree structure. Not all the
information reported in the model are useful for the different analysis, thus
only the relevant parts are parsed.

NOTE: CatBoost trains oblivious trees, i.e., trees where at each level a single
condition is checked, independently from the which node we are currently working
on. Rankeval does not exploit oblivious trees, but instead it represent them
as normal decision trees Thus the same condition will appear on all the nodes
of a single level of a tree. The reason behind this choice is to fasten the
development of the CatBoost proxy, allowing to analyze it without focusing too
much on prediction time (that is not currently measured by rankeval).
"""

import numpy as np
import logging

from .rt_ensemble import RTEnsemble


class ProxyCatBoost(object):
    """
    Class providing the implementation for loading/storing a ProxyCatBoost model
    from/to file.
    """

    @staticmethod
    def load(file_path, model):
        """
        Load the model from the file identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has been saved
        model : RTEnsemble
            The model instance to fill
        """

        try:
            import coremltools
        except ImportError:
            logging.error('Missing coremltools package!.')
            return

        coreml_model = coremltools.models.model.MLModel(file_path)

        n_trees, n_nodes = ProxyCatBoost._count_nodes(coreml_model)
        # Initialize the model and allocate the needed space
        # given the shape and size of the ensemble
        model.initialize(n_trees, n_nodes)

        n_nodes_per_tree = int(n_nodes / n_trees)

        nodes = coreml_model.get_spec().treeEnsembleRegressor.treeEnsemble.nodes
        behaviors = coremltools.proto.TreeEnsemble_pb2.TreeEnsembleParameters.\
            TreeNode.TreeNodeBehavior

        for node in nodes:
            tree_offset = node.treeId * n_nodes_per_tree
            node_id_remap = ProxyCatBoost.remap_nodeId(node.nodeId,
                                                       n_nodes_per_tree)
            node_id_off = node_id_remap + tree_offset

            if node_id_remap == 0:  # this is the root of a tree
                model.trees_root[node.treeId] = tree_offset
                model.trees_weight[node.treeId] = 1

            if node.nodeBehavior == behaviors.Value('LeafNode'):
                model.trees_nodes_value[node_id_off] = \
                    node.evaluationInfo[0].evaluationValue
            else:
                if node.nodeBehavior == behaviors.Value('BranchOnValueGreaterThan'):
                    # we need to flip the condition given we use "<="
                    left = node.falseChildNodeId
                    right = node.trueChildNodeId
                elif node.nodeBehavior == behaviors.Value('BranchOnValueLessThanEqual'):
                    right = node.falseChildNodeId
                    left = node.trueChildNodeId
                else:
                    raise AssertionError(
                        "Branching condition not supported. RankEval does not "
                        "support branching conditions different from "
                        "BranchOnValueGreaterThan or BranchOnValueLessThanEqual.")

                model.trees_nodes_value[node_id_off] = node.branchFeatureValue
                model.trees_nodes_feature[node_id_off] = node.branchFeatureIndex
                model.trees_left_child[node_id_off] = tree_offset +\
                    ProxyCatBoost.remap_nodeId(left, n_nodes_per_tree)
                model.trees_right_child[node_id_off] = tree_offset + \
                    ProxyCatBoost.remap_nodeId(right, n_nodes_per_tree)

    @staticmethod
    def remap_nodeId(nodeId, n_nodes_per_tree):
        return n_nodes_per_tree - 1 - nodeId

    @staticmethod
    def save(file_path, model):
        """
        Save the model onto the file identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has to be saved
        model : RTEnsemble
            The model RTEnsemble model to save on file

        Returns
        -------
        status : bool
            Returns true if the save is successful, false otherwise
        """
        raise NotImplementedError("Feature not implemented!")

    @staticmethod
    def _count_nodes(coreml_model):
        """
        Count the total number of nodes (both split and leaf nodes)
        in the CoreML model.

        Parameters
        ----------
        coreml_model : CoreML model
            The CoreML model to load from

        Returns
        -------
        tuple(n_trees, n_nodes) : tuple(int, int)
            The total number of trees and nodes (both split and leaf nodes)
            in the model identified by file_path.
        """

        nodes = coreml_model.get_spec().treeEnsembleRegressor.treeEnsemble.nodes

        n_trees = np.max([node.treeId for node in nodes]) + 1

        n_nodes_trees = np.empty(n_trees, dtype=np.uint16)
        for node in nodes:
            n_nodes_trees[node.treeId] = node.nodeId

        # node_Id starts from 0, thus + 1
        n_nodes = np.sum(n_nodes_trees + 1)

        return n_trees, n_nodes


# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Class providing the implementation for loading/storing a XGBoost model
from/to file. The model has to be saved using textual representation, i.e., by
using the following method:
.. code-block:: python
    import xgboost as xgb
    ...
    bst = xgb.train(param, dtrain, num_round)
    bst.dump_model('xgboost.model')

The XGBoost project is described here:
    https://github.com/dmlc/xgboost

The XGBoost format adopts a textual representation where each line of the file
represent a single split node or a leaf node, with several attributes describing
the feature and the threshold involved (in case of a split node) or the output
(in case of a leaf). Each node is identified by a unique integer as well as
additional information not usefull for rankeval and thus ignored.

NOTE: the XGBoost version 0.6 does not properly dump the model. Indeed, as
reported in the issue here:

- https://github.com/dmlc/xgboost/issues/2077

The precision of the dumping is not sufficient and cause inconsistencies with
the XGBoost model. This inconsistencies cause rankeval scoring to return
different predictions with respect to the original model. Without a fix by
XGBoost authors, DO NOT USE this proxy.
"""

import re
import sys
import numpy as np

from .rt_ensemble import RTEnsemble

tree_reg = re.compile("^booster\[(\d+)\]")
node_reg = re.compile("(\d+):\[f(\d+)<(.*)\]")
leaf_reg = re.compile("(\d+):leaf=(.+?)(,.*)?$")


class ProxyXGBoost(object):
    """
    Class providing the implementation for loading/storing a XGBoost model
    from/to file.
    """

    @staticmethod
    def load(file_path, model):
        """
        Load the model from the file identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has been saved
        model : RTEnsemble
            The model instance to fill
        """
        n_trees, n_nodes = ProxyXGBoost._count_nodes(file_path)
        # Initialize the model and allocate the needed space
        # given the shape and size of the ensemble
        model.initialize(n_trees, n_nodes)

        root_node = 0
        num_nodes = 0
        queue = list()
        with open(file_path, 'r') as f:
            for line in f:

                match_tree = tree_reg.match(line)
                if match_tree:
                    assert(len(queue) == 0)
                    curr_tree = int(match_tree.group(1))
                    root_node += num_nodes
                    num_nodes = 0
                    model.trees_root[curr_tree] = root_node
                    model.trees_weight[curr_tree] = 1
                    continue

                match_node = node_reg.search(line)
                if match_node:
                    node_id = num_nodes + root_node
                    feature_id = int(match_node.group(2).strip())
                    threshold = np.float32(match_node.group(3).strip())

                    # Needed because XGBoost use as split condition
                    # < in place of <=
                    threshold = np.nextafter(
                        threshold, threshold - 1,
                        dtype=model.trees_nodes_value.dtype)

                    model.trees_nodes_feature[node_id] = feature_id
                    model.trees_nodes_value[node_id] = threshold

                match_leaf = leaf_reg.search(line)
                if match_leaf:
                    node_id = num_nodes + root_node
                    leaf_value = float(match_leaf.group(2).strip())
                    model.trees_nodes_value[node_id] = leaf_value

                if match_node or match_leaf:
                    num_nodes += 1
                    if len(queue) > 0:
                        parent_id, child = queue.pop()
                        if child == 'L':
                            model.trees_left_child[parent_id] = node_id
                        else:
                            model.trees_right_child[parent_id] = node_id

                if match_node:
                    # two elements in the queue for the left and right children
                    # Each element is identified by a node_id and the indication
                    # of being the left or right child.
                    queue.extend([(node_id, 'R'), (node_id, 'L')])

    @staticmethod
    def save(file_path, model):
        """
        Save the model onto the file identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has to be saved
        model : RTEnsemble
            The model RTEnsemble model to save on file

        Returns
        -------
        status : bool
            Returns true if the save is successful, false otherwise
        """
        raise NotImplementedError("Feature not implemented!")

    @staticmethod
    def _count_nodes(file_path):
        """
        Count the total number of nodes (both split and leaf nodes)
        in the model identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has been saved

        Returns
        -------
        tuple(n_trees, n_nodes) : tuple(int, int)
            The total number of trees and nodes (both split and leaf nodes)
            in the model identified by file_path.
        """

        n_nodes = 0
        n_trees = 0

        with open(file_path, 'r') as f:
            for line in f:

                match = tree_reg.match(line)
                if match:
                    n_trees += 1
                    continue

                match_node = node_reg.search(line)
                if match_node:
                    n_nodes += 1

                match_leaf = leaf_reg.search(line)
                if match_leaf:
                    n_nodes += 1

        return n_trees, n_nodes



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>, 
#          Claudio Lucchese <claudio.lucchese@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
This package implements feature importance analysis.
"""

import numpy as np
from collections import deque
import xarray as xr

from ..model import RTEnsemble
from ..metrics import MSE, RMSE

from ._efficient_feature import eff_feature_importance, \
    eff_feature_importance_tree


def feature_importance(model, dataset, metric=None, normalize=True):
    """
    This method computes the feature importance relative to the given model
    and dataset.

    Parameters
    ----------
    dataset : Dataset
        The dataset used to evaluate the model (typically the one used to train
        the model).
    model : RTEnsemble
        The model whose features we want to evaluate.
    metric : rankeval.metrics.Metric
        The metric to use for compute the feature gain at each split node.
        The default metric is the Root Mean Squared Error (MSE).
    normalize : bool
        Whether to normalize the computed importance such as to have a relative
        importance (rescale importance by the value of the most important
        feature).

    Returns
    -------
    feature_importance : xarray.DataArray
        A DataArray containing the feature importance scores, one for each
        feature of the given model scored on the given dataset. Two main
        information are stored in the DataArray:
            - feature_importance: A vector of importance values, one for each
                feature in the given model. The importance values reported are
                the sum of the improvements, in terms of MSE, of each feature,
                evaluated on the given dataset. The improvements are computed as
                the delta MSE before a split node and after, evaluating how much
                the MSE is improved as a result of the split.
            - feature_count: A vector of count values, one for each feature in
                the given model. The count values reported highlights the number
                of times each feature is used in a split node, i.e., to improve
                the MSE.
    """
    if metric is None:
        metric = MSE()

    if isinstance(metric, RMSE) or isinstance(metric, MSE):
        feature_imp, feature_count = eff_feature_importance(model, dataset)
        if isinstance(metric, RMSE):
            feature_imp[0] = np.sqrt(feature_imp[0])
    else:
        # initialize features importance
        feature_imp = np.zeros(dataset.n_features, dtype=np.float32)

        # initialize features count
        feature_count = np.zeros(dataset.n_features, dtype=np.uint16)

        # default scores on the root node of the first tree
        y_pred = np.zeros(dataset.n_instances)

        # iterate trees of the model
        for tree_id in np.arange(model.n_trees):
            _feature_importance_tree(model, dataset, tree_id, y_pred, metric,
                                     feature_imp, feature_count)

    if normalize:
        feature_imp = feature_imp / feature_imp.max()

    performance = xr.DataArray([feature_imp, feature_count],
                               name='Feature Importance Analysis',
                               coords=[['importance', 'count'],
                                       np.arange(dataset.n_features, dtype=np.uint16)],
                               dims=['type', 'feature'])

    return performance


def _feature_importance_tree(model, dataset, tree_id, y_pred, metric,
                             feature_imp, feature_count):
    """
    This method computes the feature importance relative to a single tree of
    the given model.

    Parameters
    ----------
    dataset : Dataset
        The dataset used to evaluate the model (typically the one used to train
        the model).
    model : RTEnsemble
        The model whose features we want to evaluate.
    tree_id: int
        Id of the root of the tree to be evaluated.
    y_pred : numpy.array
        Current instance predictions.
    metric : rankeval.metrics.Metric
        The metric to use for compute the feature gain at each split node.
        The default metric is the Root Mean Squared Error (MSE).
    feature_imp : numpy.array
        The feature importance array, to be updated with the evaluation of the
        current tree.
    feature_ : numpy.array
        The feature count array, to be updated with the evaluation of the
        current tree.

    Returns
    -------
    y_pred_tree : numpy.array
        A vector of delta instance scores relative to the current tree.
    """
    if isinstance(metric, RMSE) or isinstance(metric, MSE):
        y_pred_tree = eff_feature_importance_tree(
            model, dataset, tree_id, y_pred, feature_imp, feature_count)
        if isinstance(metric, RMSE):
            map(np.math.sqrt, feature_imp)
        return y_pred_tree

    # The residual scores to fit
    y_target = dataset.y - y_pred

    # Set the default y_pred vector to the mean residual score
    y_pred_tree = np.full(dataset.n_instances, fill_value=y_target.mean())

    # queue with node_id to be visited and list of documents in the node
    root_node_id = model.trees_root[tree_id]
    deq = deque([(root_node_id, 0, None)])

    # we score the tree on a level basis, so that the feature_gain of each split
    # node is computed accordingly to the predictions of the parent level
    # (i.e., all the nodes in the tree with a depth < node_depth)

    # depth 0 prediction correspond to the mean residual score
    y_pred_tree_depth = [y_pred_tree]

    # support array for storing the local modification made by a split node
    y_pred_mod = np.empty(dataset.n_instances, dtype=np.float32)

    while len(deq) > 0:
        # current node info
        node_id, depth, doc_list = deq.popleft()
        feature_id = model.trees_nodes_feature[node_id]
        threshold = model.trees_nodes_value[node_id]

        feature_count[feature_id] += 1

        # this is good only for the internal nodes (non root)
        if doc_list is not None:
            left_docs = \
                doc_list[dataset.X[doc_list, feature_id] <= threshold]
            right_docs = \
                doc_list[dataset.X[doc_list, feature_id] > threshold]
        else:
            left_docs = np.where(dataset.X[:, feature_id] <= threshold)[0]
            right_docs = np.where(dataset.X[:, feature_id] > threshold)[0]

        # copy parent predictions inside y_pred_mod
        y_pred_mod[:] = y_pred_tree_depth[depth]

        # before updating leaves
        # dataset_y = y_target + y_pred
        pre_slit_metric, _ = metric.eval(dataset, y_pred_mod + y_pred)
        # ((y_target[doc_list] - y_pred_mod[doc_list]) ** 2.0).sum()

        # update y_pred including the weight of the current tree
        if left_docs.size > 0:
            y_pred_mod[left_docs] = y_target[left_docs].mean()
        if right_docs.size > 0:
            y_pred_mod[right_docs] = y_target[right_docs].mean()

        # update the y_pred of the next level
        if depth + 1 > len(y_pred_tree_depth) - 1:
            y_pred_tree_depth.append(y_pred_tree_depth[depth].copy())
        y_pred_tree_depth[depth + 1][doc_list] = y_pred_mod[doc_list]

        # after updating leaves
        post_split_metric, _ = metric.eval(dataset, y_pred_mod + y_pred)

        # compute the new metric score
        delta_metric = pre_slit_metric - post_split_metric

        # update feature importance
        feature_imp[feature_id] += delta_metric

        # if children are not leaves, add in the queue of the nodes to visit
        if not model.is_leaf_node(model.trees_left_child[node_id]):
            deq.append((model.trees_left_child[node_id], depth+1, left_docs))
        if not model.is_leaf_node(model.trees_right_child[node_id]):
            deq.append((model.trees_right_child[node_id], depth+1, right_docs))

    # update the leaves output including the tree weight
    y_pred_tree_depth[-1] *= model.trees_weight[tree_id]

    # update y_pred with the predictions made by the last level of the tree
    y_pred += y_pred_tree_depth[-1]

    return y_pred_tree_depth[-1]



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Optimized scoring of RankEval.
"""

import cython
cimport cython

# Import the Python-level symbols of numpy
import numpy as np

# Import the C-level symbols of numpy
cimport numpy as np


# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()

from cython.parallel import prange, parallel

@cython.boundscheck(False)
@cython.wraparound(False)
def basic_scoring(model, X):

    cdef np.intp_t n_instances = X.shape[0]
    cdef np.intp_t n_trees = model.n_trees

    cdef float[:, :] X_view = X
    y = np.zeros(n_instances, dtype=np.float32)
    cdef float[:] y_view = y

    cdef int[:] trees_root = model.trees_root
    cdef float[:] trees_weight = model.trees_weight
    cdef short[:] trees_nodes_feature = model.trees_nodes_feature
    cdef float[:] trees_nodes_value = model.trees_nodes_value
    cdef int[:] trees_left_child = model.trees_left_child
    cdef int[:] trees_right_child  = model.trees_right_child

    cdef int leaf_node
    cdef np.intp_t idx_tree, idx_instance
    with nogil, parallel():
        for idx_instance in prange(n_instances):
            for idx_tree in xrange(n_trees):
                leaf_node = _score_single_instance_single_tree(
                    X_view,
                    idx_instance,
                    idx_tree,
                    trees_root,
                    trees_weight,
                    trees_nodes_feature,
                    trees_nodes_value,
                    trees_left_child,
                    trees_right_child
                )

                y_view[idx_instance] += \
                    trees_nodes_value[leaf_node] * trees_weight[idx_tree]
    return y

@cython.boundscheck(False)
@cython.wraparound(False)
def detailed_scoring(model, X):

    cdef np.intp_t n_instances = X.shape[0]
    cdef np.intp_t n_trees = model.n_trees

    cdef float[:, :] X_view = X
    y_leaves = np.zeros((X.shape[0], model.n_trees), dtype=np.int32)
    cdef int[:, :] y_leaves_view = y_leaves

    partial_y = np.zeros((X.shape[0], model.n_trees), dtype=np.float32)
    cdef float[:, :] partial_y_view = partial_y

    cdef int[:] trees_root = model.trees_root
    cdef float[:] trees_weight = model.trees_weight
    cdef short[:] trees_nodes_feature = model.trees_nodes_feature
    cdef float[:] trees_nodes_value = model.trees_nodes_value
    cdef int[:] trees_left_child = model.trees_left_child
    cdef int[:] trees_right_child  = model.trees_right_child

    cdef int leaf_node
    cdef np.intp_t idx_tree, idx_instance
    with nogil, parallel():
        for idx_tree in prange(n_trees):
            for idx_instance in xrange(n_instances):
                leaf_node = _score_single_instance_single_tree(
                    X_view,
                    idx_instance,
                    idx_tree,
                    trees_root,
                    trees_weight,
                    trees_nodes_feature,
                    trees_nodes_value,
                    trees_left_child,
                    trees_right_child
                )

                y_leaves_view[idx_instance, idx_tree] = leaf_node
                partial_y_view[idx_instance, idx_tree] = \
                    trees_nodes_value[leaf_node] * trees_weight[idx_tree]

    return np.asarray(y_leaves), np.asarray(partial_y)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef int _score_single_instance_single_tree(float[:,:] X,
                                            np.intp_t idx_instance,
                                            np.intp_t idx_tree,
                                            int[:] trees_root,
                                            float[:] trees_weight,
                                            short[:] trees_nodes_feature,
                                            float[:] trees_nodes_value,
                                            int[:] trees_left_child,
                                            int[:] trees_right_child) nogil:

    # Check the usage of np.intp_t in plave of np.int16_t
    cdef int cur_node = trees_root[idx_tree]
    cdef short feature_idx
    cdef float feature_threshold
    while trees_left_child[cur_node] != -1 and trees_right_child[cur_node] != -1:
        feature_idx = trees_nodes_feature[cur_node]
        feature_threshold = trees_nodes_value[cur_node]
        if X[idx_instance, feature_idx] <= feature_threshold:
            cur_node = trees_left_child[cur_node]
        else:
            cur_node = trees_right_child[cur_node]
    return cur_node


import logging
import os
import unittest

from numpy.testing import assert_equal, assert_array_equal, \
    assert_array_almost_equal

from rankeval.dataset import Dataset
from rankeval.model import ProxyJforests
from rankeval.model import RTEnsemble
from rankeval.test.base import data_dir

model_file = os.path.join(data_dir, "Jforests.model.xml")
data_file = os.path.join(data_dir, "msn1.fold1.test.5k.txt")


class ProxyJforestsTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.model = RTEnsemble(model_file, format="Jforests")
        cls.dataset = Dataset.load(data_file, format="svmlight")

    @classmethod
    def tearDownClass(cls):
        del cls.model
        cls.model = None
        del cls.dataset
        cls.dataset = None

    def test_count_nodes(self):
        n_trees, n_nodes = ProxyJforests._count_nodes(model_file)
        assert_equal(n_trees, 2)
        assert_equal(n_nodes, 26)
        assert_equal(n_trees, self.model.trees_root.size)
        assert_equal(n_nodes, self.model.trees_nodes_value.size)

    def test_root_nodes(self):
        assert_equal((self.model.trees_root > -1).all(), True,
                     "Root nodes not set correctly")

    def test_root_nodes_adv(self):
        assert_array_equal(self.model.trees_root, [0, 13],
                           "Root nodes are not correct")

    def test_tree_weights(self):
        assert_array_almost_equal(self.model.trees_weight,
                                  [1.0, 1.0],
                                  err_msg="Tree Weights are not correct")

    def test_split_features(self):
        assert_array_equal(self.model.trees_nodes_feature,
                           [129, 129, 107, 72, 55, 54,
                            -1, -1, -1, -1, -1, -1, -1,
                            133, 72, 105, 130, 62, 121,
                            -1, -1, -1, -1, -1, -1, -1])

    def test_tree_values(self):
        assert_array_almost_equal(self.model.trees_nodes_value,
            [268.0079, 265.0144, 13.9174, 19.1123, 0.00976, 0.0185,
             -1.2156, -0.2370, -1.9329, 0.8030, -0.01019, -1.9395, 0.5840,
             0.0, 21.3979, 13.2636, 181.0142, 0.3333, -1.5976, -0.1443,
             1.3819, 1.7707, 1.7353, 0.2240, -0.3769, -1.7937],
            decimal=4,
            err_msg="Split threshold values or leaf outputs are not correct")

    def test_left_children(self):
        assert_array_equal(self.model.trees_left_child,
                           [1,  3,  5,  6,  8,  7, -1, -1, -1, -1, -1, -1, -1,
                            14, 17, 21, 22, 19, 24, -1, -1, -1, -1, -1, -1, -1])

    def test_right_children(self):
        assert_array_equal(self.model.trees_right_child,
                           [2, 4, 9, 10, 11, 12, -1, -1, -1, -1, -1, -1, -1,
                            20, 15, 16, 23, 18, 25, -1, -1, -1, -1, -1, -1, -1])

    def test_leaf_correctness(self):
        for idx, feature in enumerate(self.model.trees_nodes_feature):
            if feature == -1:
                assert_equal(self.model.trees_left_child[idx], -1,
                             "Left child of a leaf node is not empty (-1)")
                assert_equal(self.model.trees_right_child[idx], -1,
                             "Right child of a leaf node is not empty (-1)")
                assert_equal(self.model.is_leaf_node(idx), True,
                             "Leaf node not detected as a leaf")

    def test_prediction(self):
        y_pred = self.model.score(self.dataset)
        assert_array_almost_equal(y_pred[:5],
                                  [-2.083870, -1.359969, -1.359969,
                                   0.426128, -0.381351])
        assert_array_almost_equal(y_pred[-5:],
                                  [1.027176, -0.381351, -2.077223,
                                   0.658770, -0.381351])


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s',
                        level=logging.DEBUG)
    unittest.main()



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Class for efficient scoring of an ensemble-based model composed of binary regression trees on a given dataset.
"""

from ..dataset import Dataset
from ._efficient_scoring import basic_scoring, detailed_scoring


class Scorer(object):
    """
    Class for efficient scoring of an ensemble-based model composed of binary regression trees on a given dataset.

    This class can be used for simple or detailed scoring, depending on the mode selected at scoring time.
    The document scores are cached as to avoid useless re-scoring. Thus, calling multiple times the `score` method
    does not involve the scoring activity to be executed again, except for a detailed scoring following a basic scoring.
    Indeed in this situation the scoring has to be repeated as to analyze in depth the scoring behaviour.

    Parameters
    ----------
    model: RTEnsemble
        The model to use for scoring
    dataset: Dataset
        The dataset to use for scoring

    Attributes
    ----------
    model : RTEnsemble
        The model to use for scoring
    dataset : Dataset
        The dataset to use for scoring
    y_pred : numpy array of float
        The predicted scores produced by the given model for each sample of the given dataset X
    partial_y_pred : numpy 2d-array of float
        The predicted score of each tree of the model for each dataset instance

    """

    def __init__(self, model, dataset):
        self.model = model
        self.dataset = dataset

        # Save the predicted scores for each dataset instance
        self.y_pred = None

        # Save the partial scores of each tree for each dataset instance
        # (if detailed scoring is True)
        self.partial_y_pred = None

        # Save the leaf id of each tree for each dataset instance
        # (if detailed scoring is True)
        self.out_leaves = None

    def score(self, detailed):
        """

        Parameters
        ----------
        detailed : bool
            True if the class has to performs a detailed scoring, false otherwise

        Returns
        -------
        y : numpy array of float
            the predicted scores produced by the given model for each sample of the given dataset X

        Attributes
        ----------
        self.y : array of float
            The predicted scores of each dataset instance
        """

        # Skip the scoring if it has already been done (return cached results)
        if not detailed and self.y_pred is not None or \
                        detailed and self.out_leaves is not None:
            return self.y_pred

        if detailed:
            self.out_leaves, self.partial_y_pred = \
                detailed_scoring(self.model, self.dataset.X)
            self.y_pred = self.partial_y_pred.sum(axis=1)
        else:
            self.y_pred = basic_scoring(self.model, self.dataset.X)

        return self.y_pred

    def get_predicted_scores(self):
        """
        Provide an accessor to the predicted scores produced by the given model for each sample of the given dataset X

        Returns
        -------
        scores : numpy array of float
            The predicted scores produced by the given model for each sample of the given dataset X

        """
        if self.y_pred is None:
            self.score(detailed=False)
        return self.y_pred

    def get_partial_predicted_scores(self):
        """
        Provide an accessor to the partial scores produced by the given model
        for each sample of the given dataset X. Each partial score reflects the
        score produced by a single tree of the ensemble model to a single
        dataset instance. Thus, the returned numpy matrix has a shape of
        (n_instances, n_trees). The partial scores does not take into account
        the tree weights, thus for producing the final score is needed to
        multiply each row for the tree weight vector.

        Returns
        -------
        scores : numpy 2d-array of float
            The predicted score of each tree of the model for each dataset instance
        """
        if self.partial_y_pred is None:
            self.score(detailed=True)
        return self.partial_y_pred

    def get_predicted_leaves(self):
        """
        Provide an accessor to the leaves that identify the exit nodes of each
        sample of the given dataset X using the given model.

        Each leaf value reflects the output node of a single tree of the
        ensemble model to a single dataset instance. Thus, the returned numpy
        matrix has a shape of (n_instances, n_trees).

        Returns
        -------
        scores : numpy 2d-array of int
            The leaves predicted by each tree of the model on scoring
            each dataset instance.

        """
        if self.self.y_leaves is None:
            self.score(detailed=True)
        return self.self.y_leaves




""" This module implements a fast and memory-efficient (no memory copying)
loader for the svmlight / libsvm sparse dataset format.  """

# Authors: Mathieu Blondel <mathieu@mblondel.org>
#          Lars Buitinck <L.J.Buitinck@uva.nl>
# License: Simple BSD.

from ._svmlight_format import _load_svmlight_file, _dump_svmlight_file
import numpy as np
import io
import os
from contextlib import closing
from six import string_types


def _open(f, mode='r'):
    if mode not in ['r', 'w']:
        raise TypeError("expected mode to be 'r' or 'w', got %s" % mode)
    if mode == 'r':
        mode = 'rb'
    else:
        mode = 'wt'
    if isinstance(f, int):  # file descriptor
        return io.open(f, mode, closefd=False)
    elif not isinstance(f, string_types):
        raise TypeError("expected {str, int, file-like}, got %s" % type(f))

    _, ext = os.path.splitext(f)
    if ext == ".gz":
        import gzip
        return gzip.open(f, mode)
    elif ext == ".bz2":
        import bz2
        return bz2.open(f, mode)
    else:
        return open(f, mode)


def load_svmlight_file(f, query_id=False):
    """Load datasets in the svmlight / libsvm format into sparse CSR matrix

    This format is a text-based format, with one sample per line. It does
    not store zero valued features hence is suitable for sparse dataset.

    The first element of each line can be used to store a target variable
    to predict.

    This format is used as the default format for both svmlight and the
    libsvm command line programs.

    Parsing a text based source can be expensive. When working on
    repeatedly on the same dataset, it is recommended to wrap this
    loader with joblib.Memory.cache to store a memmapped backup of the
    CSR results of the first call and benefit from the near instantaneous
    loading of memmapped structures for the subsequent calls.

    Parameters
    ----------
    f : {str, file-like, int}
        (Path to) a file to load. If a path ends in ".gz" or ".bz2", it will
        be uncompressed on the fly. If an integer is passed, it is assumed to
        be a file descriptor. A file-like or file descriptor will not be closed
        by this function. A file-like object must be opened in binary mode.
    query_id : bool
        True if the query ids has to be loaded, false otherwise

    Returns
    -------
    (X, y, [query_ids])

    where X is a dense numpy matrix of shape (n_samples, n_features) and type dtype,
          y is a ndarray of shape (n_samples,).
          query_ids is a ndarray of shape(nsamples,) if query_id is True.
          Otherwise it is not returned.
    """
    if hasattr(f, "read"):
        data, labels, qids = _load_svmlight_file(f)
    else:
        with closing(_open(f)) as f:
            data, labels, qids = _load_svmlight_file(f)

    # reshape the numpy array into a matrix
    n_samples = len(labels)
    n_features = int( len(data) / n_samples )
    data.shape = (n_samples, n_features)
    if data.dtype != np.float32:
        new_data = data.astype(dtype=np.float32)
        del data
        data = new_data

    # Convert infinite values to max_float representation (SVM reader problem)
    # This patch is needed because some dataset have infinite values and because
    # the split condition is <=, while sole software uses <. In order to
    # reconduct the former condition to the latter, we slightly decrease the
    # split. However, slightly decreasing inf does not have any effect.
    data[data == np.inf] = np.finfo(data.dtype).max

    if not query_id:
        return data, labels
    else:
        return data, labels, qids


def load_svmlight_files(files, query_id=False):
    """Load dataset from multiple files in SVMlight format

    This function is equivalent to mapping load_svmlight_file over a list of
    files, except that the results are concatenated into a single, flat list
    and the samples vectors are constrained to all have the same number of
    features.

    Parameters
    ----------
    files : iterable over {str, file-like, int}
        (Paths of) files to load. If a path ends in ".gz" or ".bz2", it will
        be uncompressed on the fly. If an integer is passed, it is assumed to
        be a file descriptor. File-likes and file descriptors will not be
        closed by this function. File-like objects must be opened in binary
        mode

    n_features: int or None
        The number of features to use. If None, it will be inferred from the
        first file. This argument is useful to load several files that are
        subsets of a bigger sliced dataset: each subset might not have
        examples of every feature, hence the inferred shape might vary from
        one slice to another.

    Returns
    -------
    [X1, y1, ..., Xn, yn]

    where each (Xi, yi, [comment_i, query_id_i]) tuple is the result from
    load_svmlight_file(files[i]).

    Rationale
    ---------
    When fitting a model to a matrix X_train and evaluating it against a
    matrix X_test, it is essential that X_train and X_test have the same
    number of features (X_train.shape[1] == X_test.shape[1]). This may not
    be the case if you load them with load_svmlight_file separately.

    See also
    --------
    load_svmlight_file
    """
    files = iter(files)
    result = list(load_svmlight_file(next(files), query_id=query_id))

    for f in files:
        result += load_svmlight_file(f, query_id=query_id)

    return result


def dump_svmlight_file(X, y, f, query_id=None, zero_based=False):
    """Dump the dataset in svmlight / libsvm file format.

    This format is a text-based format, with one sample per line. It does
    not store zero valued features hence is suitable for sparse dataset.

    The first element of each line can be used to store a target variable
    to predict.

    Parameters
    ----------
    X : CSR sparse matrix, shape = [n_samples, n_features]
        Training vectors, where n_samples is the number of samples and
        n_features is the number of features.

    y : array-like, shape = [n_samples]
        Target values.

    f : {str, file-like, int}
        (Path to) a file to dump. If a path ends in ".gz" or ".bz2", it will
        be compressed on the fly. If an integer is passed, it is assumed
        to be a file descriptor. A file-like or file descriptor will not be
        closed by this function. A file-like object must be opened in
        text mode.

    comment : list, optional 
        Comments to append to each row after a # character
        If specified, len(comment) must equal n_samples

    query_id: list, optional 
        Query identifiers to prepend to each row
        If specified, len(query_id) must equal n_samples

    zero_based : boolean, optional
        Whether column indices should be written zero-based (True) or one-based
        (False). Default is False.
    """
    X = np.array(X, dtype=np.float32)
    y = np.array(y, dtype=np.float32)
    if query_id is None:
        query_id = np.ndarray(0, dtype=np.int32)
    else:
        query_id = np.array(query_id, dtype=np.int32)

    if hasattr(f, "write"):
        _dump_svmlight_file(f, X, y, query_id, int(zero_based))
    else:
        with closing(_open(f, mode='w')) as f:
            _dump_svmlight_file(f, X, y, query_id, int(zero_based))


# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Cristina Muntean <cristina.muntean@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import numpy as np
from rankeval.metrics.metric import Metric


class Recall(Metric):
    """
    This class implements Recall as:
    (relevant docs & retrieved docs) / relevant docs.

    It allows setting custom values for cutoff and threshold, otherwise it uses
    the default values.

    """

    _threshold = 1

    def __init__(self, name='R', no_relevant_results=0.0,
                 cutoff=None, threshold=_threshold):
        """
        This is the constructor of Recall, an object of type Metric, with
        the name R. The constructor also allows setting custom values
        for cutoff and threshold, otherwise it uses the default values.

        Parameters
        ----------
        name: string
            R
        no_relevant_results: float
            Float indicating how to treat the cases where then are no relevant
            results (e.g. 0.0).
        cutoff: int
            The top k results to be considered at per query level (e.g. 10)
        threshold: float
            This parameter considers relevant results all instances with labels
            different from 0, thus with a minimum label value of 1. It can be
            set to other values as well (e.g. 3), in the range of possible
            labels.
        """
        super(Recall, self).__init__(name)
        self.no_relevant_results = no_relevant_results
        self.cutoff = cutoff
        self.threshold = threshold

    def eval(self, dataset, y_pred):
        """
        This method computes the Recall score over the entire dataset and the
        detailed scores per query. It calls the eval_per query method for each
        query in order to get the detailed Recall score.

        Parameters
        ----------
        dataset : Dataset
            Represents the Dataset object on which to apply Recall.
        y_pred : numpy 1d array of float
            Represents the predicted document scores for each instance in the
            dataset.

        Returns
        -------
        avg_score: float
            The overall Recall score (averages over the detailed precision
            scores).
        detailed_scores: numpy 1d array of floats
            The detailed Recall scores for each query, an array of length of
            the number of queries.
        """
        return super(Recall, self).eval(dataset, y_pred)


    def eval_per_query(self, y, y_pred):
        """
        This methods computes Recall at per query level (on the instances
        belonging to a specific query). The Recall per query is calculated as
        <(relevant docs & retrieved docs) / relevant docs>.

        Parameters
        ----------
        y: numpy array
            Represents the labels of instances corresponding to one query in
            the dataset (ground truth).
        y_pred: numpy array.
            Represents the predicted document scores obtained during the model
            scoring phase for that query.

        Returns
        -------
        recall: float
            The Recall score per query.

        """
        idx_y_pred_sorted = np.argsort(y_pred)[::-1]
        if self.cutoff is not None:
            idx_y_pred_sorted = idx_y_pred_sorted[:self.cutoff]

        n_relevant_retrieved = (y[idx_y_pred_sorted] >= self.threshold).sum()
        n_relevant = (y >= self.threshold).sum()

        if n_relevant != 0:
            return float(n_relevant_retrieved) / n_relevant
        else:
            return self.no_relevant_results

    def __str__(self):
        s = self.name
        if self.cutoff is not None:
            s += "@{}".format(self.cutoff)
        if self.threshold != self._threshold:
            s += "[>{}]".format(self.threshold)
        return s



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Cristina Muntean <cristina.muntean@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from collections import defaultdict
import numpy as np

from rankeval.metrics.dcg import DCG
from rankeval.metrics.metric import Metric


class NDCG(Metric):
    """
    This class implements NDCG with several parameters.

    """

    def __init__(self, name='NDCG', cutoff=None, no_relevant_results=1.0,
                 implementation="exp"):
        """
        This is the constructor of NDCG, an object of type Metric, with the
        name NDCG.
        The constructor also allows setting custom values
            - cutoff: the top k results to be considered at per query level
            - no_relevant_results: is a float values indicating how to treat
                the cases where then are no relevant results
            - ties: indicates how we should consider the ties
            - implementation: indicates whether to consider the flat or the
                exponential NDCG formula

        Parameters
        ----------
        name: string
            NDCG
        cutoff: int
            The top k results to be considered at per query level (e.g. 10)
        no_relevant_results: float
            Float indicating how to treat the cases where then are no relevant
            results (e.g. 0.5). Default is 1.0.
        implementation: string
            Indicates whether to consider the flat or the exponential DCG
            formula: "flat" or "exp" (default).
        """

        super(self.__class__, self).__init__(name)
        self.cutoff = cutoff
        self.no_relevant_results = no_relevant_results
        self.implementation = implementation
        self.dcg = DCG(cutoff=self.cutoff,
                       implementation=self.implementation)

        self._current_dataset = None
        self._current_rel_qid = None
        self._cache_idcg_score = defaultdict(int)

    def eval(self, dataset, y_pred):
        """
        The method computes NDCG by taking as input the dataset and the
        predicted document scores (obtained with the scoring methods). It
        returns the averaged NDCG score over the entire dataset and the
        detailed NDCG scores per query.

        Parameters
        ----------
        dataset : Dataset
            Represents the Dataset object on which to apply NDCG.
        y_pred : numpy 1d array of float
            Represents the predicted document scores for each instance in the
            dataset.

        Returns
        -------
        avg_score: float
            Represents the average NDCG over all NDCG scores per query.
        detailed_scores: numpy array of floats
            Represents the detailed NDCG scores for each query. It has the
            length of n_queries.

        """
        # used to cache ideal DCG scores on a dataset basis
        self._current_dataset = dataset
        self._current_rel_qid = 0

        # Compute the ideal DCG scores only once and cache them
        if self._current_dataset not in self._cache_idcg_score:

            idcg_score = np.ndarray(shape=dataset.n_queries, dtype=np.float32)
            for rel_id, (qid, q_y, _) in enumerate(
                    self.query_iterator(dataset, dataset.y)):
                idcg_score[rel_id] = self.dcg.eval_per_query(q_y, q_y)

            self._cache_idcg_score[self._current_dataset] = idcg_score

        return super(self.__class__, self).eval(dataset, y_pred)

    def eval_per_query(self, y, y_pred):
        """
        This method helps compute the NDCG score per query. It is called by the
        eval function which averages and aggregates the scores for each query.

        It calculates NDCG per query as <dcg_score/idcg_score>.
        If there are no relevant results, NDCG returns the values set by default
        or by the user when creating the metric.

        Parameters
        ----------
        y: numpy array
            Represents the labels of instances corresponding to one query in the
            dataset (ground truth).
        y_pred: numpy array.
            Represents the predicted document scores obtained during the model
            scoring phase for that query.

        Returns
        -------
        dcg: float
            Represents the DCG score for one query.
        """
        dcg_score = self.dcg.eval_per_query(y, y_pred)

        if self._current_rel_qid is not None:
            idcg_score = \
                self._cache_idcg_score[self._current_dataset][self._current_rel_qid]
            self._current_rel_qid += 1
        else:
            idcg_score = self.dcg.eval_per_query(y, y)

        if idcg_score != 0:
            ndcg = dcg_score / idcg_score
        else:
            ndcg = self.no_relevant_results
        return ndcg

    def __str__(self):
        s = self.name
        if self.cutoff is not None:
            s += "@{}".format(self.cutoff)
        return s











# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Class providing the implementation for loading/storing a QuickRank model
from/to file.

The Jforests project is described here: https://github.com/yasserg/jforests

The Jforests format adopts an XML representation. There is an ensemble node,
with a sub-node for each tree, identified by the "Tree" tag, followed by the
description of the tree (with splitting and leaf nodes). The splitting nodes are
described with two information: the feature-id used for splitting, and the
threshold value. Leaf nodes on the other hand are described by a "LeafOutputs"
tag with the value as content.
"""

from .rt_ensemble import RTEnsemble

try:
    import xml.etree.cElementTree as etree
except ImportError:
    import xml.etree.ElementTree as etree


class ProxyJforests(object):
    """
    Class providing the implementation for loading/storing a Jforests model
    from/to file.
    """

    @staticmethod
    def load(file_path, model):
        """
        Load the model from the file identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has been saved
        model : RTEnsemble
            The model instance to fill
        """
        n_trees, n_nodes = ProxyJforests._count_nodes(file_path)
        # Initialize the model and allocate the needed space
        # given the shape and size of the ensemble
        model.initialize(n_trees, n_nodes)

        # get an iterable
        context = etree.iterparse(file_path, events=("start", "end"))

        # get the root element
        _, root = next(context)

        curr_tree = -1
        root_node = 0
        num_leaves = num_splits = 0

        for event, elem in context:

            if event == 'start' and elem.tag == 'Tree':
                    curr_tree += 1  # increase the current number index
                    root_node += num_leaves + num_splits
                    # save the curr node as the root of a new tree
                    model.trees_root[curr_tree] = root_node
                    model.trees_weight[curr_tree] = elem.attrib['weight']

            if event == 'end':

                if elem.tag == 'SplitFeatures':
                    split_features = map(int, elem.text.split(" "))
                    num_splits = 0
                    for pos, feature in enumerate(split_features):
                        num_splits += 1
                        model.trees_nodes_feature[root_node + pos] = feature
                elif elem.tag == 'LeftChildren':
                    left_children = map(int, elem.text.split(" "))
                    for pos, child in enumerate(left_children):
                        if child >= 0:
                            model.trees_left_child[root_node + pos] = \
                                root_node + child
                        else:
                            model.trees_left_child[root_node + pos] = \
                                root_node + num_splits + abs(child) - 1
                elif elem.tag == 'RightChildren':
                    right_children = map(int, elem.text.split(" "))
                    for pos, child in enumerate(right_children):
                        if child >= 0:
                            model.trees_right_child[root_node + pos] = \
                                root_node + child
                        else:
                            model.trees_right_child[root_node + pos] = \
                                root_node + num_splits + abs(child) - 1
                elif elem.tag == 'OriginalThresholds':
                    thresholds = map(float, elem.text.split(" "))
                    for pos, threshold in enumerate(thresholds):
                        model.trees_nodes_value[root_node + pos] = threshold
                elif elem.tag == 'LeafOutputs':
                    leaf_values = map(float, elem.text.split(" "))
                    num_leaves = 0
                    for pos, leaf_value in enumerate(leaf_values):
                        num_leaves += 1
                        model.trees_nodes_value[root_node + num_splits + pos] \
                            = leaf_value

            # clear the memory
            if event == 'end':
                elem.clear()    # discard the element
                root.clear()    # remove child reference from the root

    @staticmethod
    def save(file_path, model):
        """
        Save the model onto the file identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has to be saved
        model : RTEnsemble
            The model RTEnsemble model to save on file

        Returns
        -------
        status : bool
            Returns true if the save is successful, false otherwise
        """
        raise NotImplementedError("Feature not implemented!")

    @staticmethod
    def _count_nodes(file_path):
        """
        Count the total number of nodes (both split and leaf nodes)
        in the model identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has been saved

        Returns
        -------
        tuple(n_trees, n_nodes) : tuple(int, int)
            The total number of trees and nodes (both split and leaf nodes)
            in the model identified by file_path.
        """
        # get an iterable
        # NOTE: it seems like there is a bug inside lxmx since selecting only
        # terminal tags with events=("end",) some tags are skipped...
        context = etree.iterparse(file_path, events=("start", "end"))

        # get the root element
        _, root = next(context)

        n_nodes = 0
        n_trees = 0
        for event, elem in context:
            if event != "end":
                continue
            if elem.tag == 'Tree':
                n_trees += 1
            elif elem.tag == 'SplitFeatures' or elem.tag == 'LeafOutputs':
                n_nodes += len(elem.text.split(" "))

            elem.clear()    # discard the element
            root.clear()    # remove root reference to the child

        return n_trees, n_nodes



"""
This file implements the feature importance analysis in an efficient way. The
limit is that the metric used to compute the gain for each split is hardcoded
in the source code and is the MSE.
"""

import cython
cimport cython

# Import the Python-level symbols of numpy
import numpy as np

# Import the C-level symbols of numpy
cimport numpy as np

# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()

cdef extern from "_efficient_feature_impl.h":
    void c_feature_importance(
        const float* X,
        const float* y,
        const int* trees_root,
        const float* trees_weight,
        const short* trees_nodes_feature,
        const float* trees_nodes_value,
        const int* trees_left_child,
        const int* trees_right_child,
        float* feature_imp,
        short* feature_count,
        const int n_instances,
        const int n_features,
        const int n_trees);

    void c_feature_importance_tree(
        const float* X,
        const float* y,
        const int* trees_root,
        const float* trees_weight,
        const short* trees_nodes_feature,
        const float* trees_nodes_value,
        const int* trees_left_child,
        const int* trees_right_child,
        const int tree_id,
        float* feature_imp,
        short* feature_count,
        const int n_instances,
        const int n_features,
        float* y_pred,
        float* y_pred_tree);

@cython.boundscheck(False)
@cython.wraparound(False)
def eff_feature_importance(model, dataset):

    # initialize features importance
    feature_imp = np.zeros(dataset.n_features, dtype=np.float32)

    # initialize features importance
    feature_count = np.zeros(dataset.n_features, dtype=np.uint16)

    c_feature_importance(
        <float*> np.PyArray_DATA(dataset.X),
        <float*> np.PyArray_DATA(dataset.y),
        <int*> np.PyArray_DATA(model.trees_root),
        <float*> np.PyArray_DATA(model.trees_weight),
        <short*> np.PyArray_DATA(model.trees_nodes_feature),
        <float*> np.PyArray_DATA(model.trees_nodes_value),
        <int*> np.PyArray_DATA(model.trees_left_child),
        <int*> np.PyArray_DATA(model.trees_right_child),
        <float*> np.PyArray_DATA(feature_imp),
        <short*> np.PyArray_DATA(feature_count),
        dataset.X.shape[0],
        dataset.X.shape[1],
        model.n_trees);

    return np.asarray(feature_imp, dtype=np.float32), \
           np.asarray(feature_count, dtype=np.uint16)

@cython.boundscheck(False)
@cython.wraparound(False)
def eff_feature_importance_tree(model, dataset, tree_id, y_pred,
                             feature_imp, feature_count):

    y_pred_tree = np.zeros(dataset.n_instances, dtype=np.float32);

    c_feature_importance_tree(
        <float*> np.PyArray_DATA(dataset.X),
        <float*> np.PyArray_DATA(dataset.y),
        <int*> np.PyArray_DATA(model.trees_root),
        <float*> np.PyArray_DATA(model.trees_weight),
        <short*> np.PyArray_DATA(model.trees_nodes_feature),
        <float*> np.PyArray_DATA(model.trees_nodes_value),
        <int*> np.PyArray_DATA(model.trees_left_child),
        <int*> np.PyArray_DATA(model.trees_right_child),
        tree_id,
        <float*> np.PyArray_DATA(feature_imp),
        <short*> np.PyArray_DATA(feature_count),
        dataset.X.shape[0],
        dataset.X.shape[1],
        <float*> np.PyArray_DATA(y_pred),
        <float*> np.PyArray_DATA(y_pred_tree));

    return np.asarray(y_pred_tree, dtype=np.float32)


# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Class providing the implementation for loading/storing a LightGBM model
from/to file.

The LightGBM project is described here:
    https://github.com/Microsoft/LightGBM

The LightGBM format adopts a textual representation using arrays for storing
split nodes (both features and thresholds), leaf values and tree structure.
Not all the information reported in the model are useful for the different
analysis, thus only the relevant parts are parsed.

NOTE: the leaves output of the regression trees already take into account the
weight of the tree (i.e., the learning rate or shrinkage factor). In order to
maintain the scoring made by rankeval (that multiply the leaf output by the tree
weight), the weight of the trees have been set equals to 1.

NOTE: currently rankeval does not support missing features. This features is not
strictly required because all the most popular open-domain Learning to Rank
datasets do not have missing values. To this regard, take thus into account that
you may experience an unexpected behaviour if the LightGBM model you are trying
to load is trained on a dataset with missing values (especially in the case you
use the setting zero_as_missing=true).
"""

import re

import numpy as np

from .rt_ensemble import RTEnsemble

tree_reg = re.compile("^Tree=(\d+)")
num_leaves_reg = re.compile("^num_leaves=(\d+)")
split_feature_reg = re.compile("^split_feature=(.*)")
threshold_reg = re.compile("^threshold=(.*)")
decision_type_reg = re.compile("^decision_type=(.*)")
default_value_reg = re.compile("^default_value=(.*)")
left_child_reg = re.compile("^left_child=(.*)")
right_child_reg = re.compile("^right_child=(.*)")
leaf_parent_reg = re.compile("^leaf_parent=(.*)")
leaf_value_reg = re.compile("^leaf_value=(.*)")
shrinkage_reg = re.compile("^shrinkage=(.*)")
has_categorical_reg = re.compile("^has_categorical=(.*)")


class ProxyLightGBM(object):
    """
    Class providing the implementation for loading/storing a LightGBM model
    from/to file.
    """

    @staticmethod
    def load(file_path, model):
        """
        Load the model from the file identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has been saved
        model : RTEnsemble
            The model instance to fill
        """
        n_trees, n_nodes = ProxyLightGBM._count_nodes(file_path)
        # Initialize the model and allocate the needed space
        # given the shape and size of the ensemble
        model.initialize(n_trees, n_nodes)

        curr_tree = root_node = 0
        num_leaves = num_splits = 0
        with open(file_path, 'r') as f:
            for line in f:

                match = tree_reg.match(line)
                if match:
                    curr_tree = int(match.group(1))
                    root_node += num_leaves + num_splits
                    model.trees_root[curr_tree] = root_node
                    continue

                match = split_feature_reg.match(line)
                if match:
                    split_features = list( map(int, match.group(1).strip().split()) )
                    for pos, feature in enumerate(split_features):
                        model.trees_nodes_feature[root_node + pos] = feature
                    num_splits = len(split_features)
                    continue

                match = threshold_reg.match(line)
                if match:
                    thresholds = list( map(float, match.group(1).strip().split()) )
                    for pos, threshold in enumerate(thresholds):
                        model.trees_nodes_value[root_node + pos] = threshold
                    continue

                match = left_child_reg.match(line)
                if match:
                    left_children = list( map(int, match.group(1).strip().split()) )
                    for pos, child in enumerate(left_children):
                        if child >= 0:
                            model.trees_left_child[root_node + pos] = \
                                root_node + child
                        else:
                            model.trees_left_child[root_node + pos] = \
                                root_node + num_splits + abs(child) - 1
                    continue

                match = right_child_reg.match(line)
                if match:
                    right_children = list( map(int, match.group(1).strip().split()) )
                    for pos, child in enumerate(right_children):
                        if child >= 0:
                            model.trees_right_child[root_node + pos] = \
                                root_node + child
                        else:
                            model.trees_right_child[root_node + pos] = \
                                root_node + num_splits + abs(child) - 1
                    continue

                match = shrinkage_reg.match(line)
                if match:
                    # weight should be the shrinkage but it is set to 1 because
                    # leaves output already take into account the shrinkage
                    # shrinkage = float(match.group(1))
                    model.trees_weight[curr_tree] = 1.0
                    continue

                match = leaf_value_reg.match(line)
                if match:
                    leaf_values = list( map(float, match.group(1).strip().split()) )
                    num_leaves = len(leaf_values)
                    for pos, leaf_value in enumerate(leaf_values):
                        model.trees_nodes_value[root_node + num_splits + pos] \
                            = leaf_value
                    num_splits = len(split_features)
                    continue

                match = decision_type_reg.match(line)
                if match:
                    types = np.array(match.group(1).strip().split(), dtype=int)
                    if (types != 2).any():
                        raise AssertionError("Decision Tree not supported. RankEval does not "
                                             "support categorical features and missing values.")
                    continue

                match = default_value_reg.match(line)
                if match:
                    values = np.array(match.group(1).strip().split(),
                                      dtype=np.float64)
                    if values.any():
                        raise AssertionError("Missing Values not supported!")
                    continue

                match = has_categorical_reg.match(line)
                if match:
                    categorical = bool(int(match.group(1)))
                    if categorical:
                        raise AssertionError("Decision Tree not supported")
                    continue


    @staticmethod
    def save(file_path, model):
        """
        Save the model onto the file identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has to be saved
        model : RTEnsemble
            The model RTEnsemble model to save on file

        Returns
        -------
        status : bool
            Returns true if the save is successful, false otherwise
        """
        raise NotImplementedError("Feature not implemented!")

    @staticmethod
    def _count_nodes(file_path):
        """
        Count the total number of nodes (both split and leaf nodes)
        in the model identified by file_path.

        Parameters
        ----------
        file_path : str
            The path to the filename where the model has been saved

        Returns
        -------
        tuple(n_trees, n_nodes) : tuple(int, int)
            The total number of trees and nodes (both split and leaf nodes)
            in the model identified by file_path.
        """

        n_nodes = 0
        n_trees = 0

        with open(file_path, 'r') as f:
            for line in f:
                match = num_leaves_reg.match(line)
                if match:
                    n_nodes += int(match.group(1))
                    continue
                match = split_feature_reg.match(line)
                if match:
                    n_nodes += len(match.group(1).strip().split())
                    continue
                match = tree_reg.match(line)
                if match:
                    n_trees += 1
                    continue

        return n_trees, n_nodes



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Franco Maria Nardini <francomaria.nardini@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

class DatasetContainer(object):
    """
    This class is a container used to easily manage a dataset and associated
    learning to rank models trained by using it. It also offers the possibility
    to store the license coming with public dataset.
    """
    train_dataset = None
    validation_dataset = None
    test_dataset = None

    license_agreement = ''

    model_filenames = None


# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Cristina Muntean <cristina.muntean@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


import numpy as np
from rankeval.metrics.metric import Metric


class MAP(Metric):
    """
    This class implements MAP with several parameters. We implemented MAP as in
    https://www.kaggle.com/wiki/MeanAveragePrecision, adapted from:
    http://en.wikipedia.org/wiki/Information_retrieval
    https://www.ethz.ch/content/dam/ethz/special-interest/gess/computational-social-science-dam/documents/education/Spring2017/ML/LinkPrediction.pdf
    http://sdsawtelle.github.io/blog/output/mean-average-precision-MAP-for-recommender-systems.html
    """

    def __init__(self, name='MAP', cutoff=None, no_relevant_results=1.0):
        """
        This is the constructor of MAP, an object of type Metric, with
        the name MAP. The constructor also allows setting custom values in the
        following parameters.

        Parameters
        ----------
        name: string
            MAP
        cutoff: int
            The top k results to be considered at per query level (e.g. 10),
            otherwise the default value is None and is computed on all the
            instances of a query.
        no_relevant_results: float
            Float indicating how to treat the cases where then are no relevant
            results (e.g. 0.5). Default is 1.0.
        """
        super(MAP, self).__init__(name)
        self.cutoff = cutoff
        self.no_relevant_results = no_relevant_results

    def eval(self, dataset, y_pred):
        """
        This method takes the AP@k for each query and calculates the average,
        thus MAP@k.

        Parameters
        ----------
        dataset : Dataset
            Represents the Dataset object on which to apply MAP.
        y_pred : numpy 1d array of float
            Represents the predicted document scores for each instance in
            the dataset.

        Returns
        -------
        avg_score: float
            The overall MAP@k score (averages over the detailed MAP scores).
        detailed_scores: numpy 1d array of floats
            The detailed AP@k scores for each query, an array of length of
            the number of queries.
        """
        return super(MAP, self).eval(dataset, y_pred)

    def eval_per_query(self, y, y_pred):
        """
        This methods computes AP@k at per query level (on the instances
        belonging to a specific query). The AP@k per query is calculated as

        ap@k = sum( P(k) / min(m,n) ), for k=1,n

        where:
            - P(k) means the precision at cut-off k in the item list. P(k)
            equals 0 when the k-th item is not followed upon recommendation
            - m is the overall number of relevant documents
            - n is the number of predicted documents

        If the denominator is zero, P(k)/min(m,n) is set to zero.

        Parameters
        ----------
        y: numpy array
            Represents the labels of instances corresponding to one query in
            the dataset (ground truth).
        y_pred: numpy array.
            Represents the predicted document scores obtained during the model
            scoring phase for that query.

        Returns
        -------
        map : float
            The MAP per query.
        """
        idx_y_pred_sorted = np.argsort(y_pred)[::-1]
        if self.cutoff is not None:
            idx_y_pred_sorted = idx_y_pred_sorted[:self.cutoff]

        n_retrieved = len(idx_y_pred_sorted)
        precision_at_i = 0.
        n_relevant_retrieved_at_i = 0.
        for i in range(n_retrieved):
            if y[idx_y_pred_sorted[i]] > 0:
                n_relevant_retrieved_at_i += 1
                precision_at_i += n_relevant_retrieved_at_i / (i + 1)

        if n_relevant_retrieved_at_i > 0:
            return precision_at_i / min(n_retrieved, np.count_nonzero(y))
        else:
            return self.no_relevant_results

    def __str__(self):
        s = self.name
        if self.cutoff is not None:
            s += "@{}".format(self.cutoff)
        return s



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import numpy as np
from rankeval.metrics.metric import Metric


class PrecisionMax(Metric):
    """
    This class implements Precision as:
    (relevant docs & retrieved docs) / retrieved docs.

    The particularity of this implementation is that the threshold is not global
    to all the queries as in `Precision`, but is dynamically computed for every
    query considering the document with top label. Thus this metric enables to
    compute the precision regardless of a fixed label.
    """

    def __init__(self, name='P', cutoff=None):
        """
        This is the constructor of Precision, an object of type Metric, with
        the name P. The constructor also allows setting custom values for cutoff
        and threshold, otherwise it uses the default values.

        Parameters
        ----------
        name: string
            P
        cutoff: int
            The top k results to be considered at per query level (e.g. 10)

        """
        super(PrecisionMax, self).__init__(name)
        self.cutoff = cutoff

    def eval(self, dataset, y_pred):
        """
        This method computes the Precision score over the entire dataset and
        the detailed scores per query. It calls the eval_per query method for
        each query in order to get the detailed Precision score.

        Parameters
        ----------
        dataset : Dataset
            Represents the Dataset object on which to apply Precision.
        y_pred : numpy 1d array of float
            Represents the predicted document scores for each instance in the
            dataset.

        Returns
        -------
        avg_score: float
            The overall Precision score (averages over the detailed precision
            scores).
        detailed_scores: numpy 1d array of floats
            The detailed Precision scores for each query, an array of length of
            the number of queries.
        """
        return super(PrecisionMax, self).eval(dataset, y_pred)

    def eval_per_query(self, y, y_pred):
        """
        This methods computes Precision at per query level (on the instances
        belonging to a specific query). The Precision per query is calculated as
        <(relevant docs & retrieved docs) / retrieved docs>.

        Parameters
        ----------
        y: numpy array
            Represents the labels of instances corresponding to one query in
            the dataset (ground truth).
        y_pred: numpy array.
            Represents the predicted document scores obtained during the model
            scoring phase for that query.

        Returns
        -------
        precision: float
            The precision per query.
        """
        idx_y_pred_sorted = np.argsort(y_pred)[::-1]
        if self.cutoff is not None:
            idx_y_pred_sorted = idx_y_pred_sorted[:self.cutoff]

        n_relevant_retrieved = (y[idx_y_pred_sorted] >= y.max()).sum()
        return float(n_relevant_retrieved) / len(idx_y_pred_sorted)

    def __str__(self):
        s = self.name
        if self.cutoff is not None:
            s += "@{}".format(self.cutoff)
        return s



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors:  Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


"""
This package implements several topological analysis focused on the
topological characteristics of ensemble-based LtR models. These
functionalities can be applied to several models,
so as to have a direct comparison of the shape of the resulting
forests (e.g., trained by different LtR algorithms).
"""

import numpy as np
import scipy.stats

from ..model import RTEnsemble
from ._efficient_topological import efficient_topological_analysis


def topological_analysis(model, include_leaves=True):
    """
    This method implements the topological analysis of a ensemble-based
    LtR model. Given a model, it studies the shape of each tree composing
    the model and return several information useful for having insights
    about the shape of the trees, their completeness (level by level) as
    well as min/max/mean height and the fraction of trees having a specific
    node (where each node is identified by a pair of coordinates row-col,
    with row highlighting the depth and col the column with respect to a
    full binary tree).

    Parameters
    ----------
    model : RTEnsemble
        The model to analyze
    include_leaves : bool
        Whether the leaves has to be included in the analysis or not

    Returns
    -------
    object : TopologicalAnalysisResult
        The topological result, to use for retrieving several information
    """
    return TopologicalAnalysisResult(model, include_leaves)


class TopologicalAnalysisResult(object):
    """
    This class is used to return the topological analysis made on the model.
    Several low-level information are stored in this class, and then
    re-elaborated to provide high-level analysis.
    """

    def __init__(self, model, include_leaves):
        """
        Analyze the model in a topological perspective

        Parameters
        ----------
        model : RTEnsemble
            the model to analyze from the topological perspective
        include_leaves : bool
            Whether the leaves has to be included in the analysis or not

        Attributes
        ----------
        model : RTEnsemble
            The model analyzed
        height_trees : numpy array
            The ordered height of each trees composing the ensemble
        topology : scipy.sparse.csr_matrix
            The matrix used to store low-level information related to the
            aggregated shape of the trees. Each matrix cell identifies a
            tree node with a pair of coordinates row-col, with row
            highlighting the depth and col the column with respect
            to a full binary tree.
        """
        self.model = model
        self.topology, self.height_trees = efficient_topological_analysis(model, include_leaves)

    def describe_tree_height(self):
        """
        Computes several descriptive statistics of the height of the trees.

        Returns
        -------
        nobs : int
           Number of trees
        minmax: tuple of ndarrays or floats
           Minimum and maximum height of trees
        mean : ndarray or float
           Arithmetic mean of tree heights.
        variance : ndarray or float
           Unbiased variance of the tree heights.
           denominator is number of trees minus one.
        skewness : ndarray or float
           Skewness, based on moment calculations with denominator equal to
           the number of trees, i.e. no degrees of freedom correction.
        kurtosis : ndarray or float
           Kurtosis (Fisher).  The kurtosis is normalized so that it is
           zero for the normal distribution.  No degrees of freedom are used.
        """
        return scipy.stats.describe(self.height_trees)

    def avg_tree_shape(self):
        """
        Computes the fraction of trees having each node with respect to a
        full binary tree. The fraction is obtained by normalizing the count
        by the number of trees composing the ensemble model.

        Returns
        -------
        fractions : scipy.sparse.csr_matrix
            Sparse matrix with the same shape of the topology matrix, where
            each matrix cell identifies a tree node by a pair of coordinates
            row-col, with row highlighting the depth and col the column with
            respect to a full binary tree. Each cell value highlights how many
            trees have the specific node, normalized by the number of trees.
        """
        return self.topology / self.model.n_trees

    def fullness_per_level(self):
        """
        Computes the normalized number of trees with full level i, for each
        level of a full binary tree. The normalization is done by the number
        of trees.

        Returns
        -------
        fullness : np.array
            An array long as the maximum height of a tree in the ensemble, and
            where the j-th cell highlight how much the j-th level of the trees
            is full (normalized by the number of trees).
        """
        # Row-sums are directly supported, and the structure of the CSR format means that
        # the difference between successive values in the indptr array correspond exactly
        # to the number of nonzero elements in each row.
        sums = self.topology.sum(axis=1).A1
        counts = np.diff(self.topology.indptr)
        return sums / counts / self.model.n_trees



import logging
import os
import unittest

import numpy as np
from nose.tools import raises
from numpy.testing import assert_equal, assert_array_equal, \
    assert_array_almost_equal

try:
    from sklearn.datasets import load_svmlight_file as sk_load_svmlight_file
    scikit_missing = False
except ImportError:
    scikit_missing = True

from rankeval.dataset.svmlight_format import load_svmlight_file, \
    load_svmlight_files, dump_svmlight_file
from rankeval.test.base import data_dir

datafile = os.path.join(data_dir, "svmlight_classification.txt")
invalidfile = os.path.join(data_dir, "svmlight_invalid.txt")
qid_datafile = os.path.join(data_dir, "svmlight_classification_qid.txt")


class SVMLightLoaderTestCase(unittest.TestCase):

    def test_load_svmlight_qid_file(self):
        X, y, q = load_svmlight_file(qid_datafile, query_id=True)

        # test X's shape
        assert_array_equal(X.shape, (4, 33))
        #print X

        # test X's non-zero values
        # tests X's zero values
        # test can change X's values

        # test y
        assert_array_equal(y, [1, 2, 0, 3])

        # test q
        # print q
        assert_array_equal(q, [1, 37, 37, 12])

    def test_load_svmlight_file_empty_qid(self):
        X, y, q = load_svmlight_file(datafile, query_id=True)

        # test X's shape
        assert_array_equal(X.shape, (3, 20))

        # test X's non-zero values
        # tests X's zero values
        # test can change X's values

        # test y
        assert_array_equal(y, [1, 2, 3])

        # test q
        assert_equal(q.shape[0], 0)

    def test_load_svmlight_file(self):
        X, y = load_svmlight_file(datafile)

        # test X's shape
        assert_array_equal(X.shape, (3, 20))

        # test X's non-zero values
        # tests X's zero values
        # test can change X's values

        # test y
        assert_array_equal(y, [1, 2, 3])

    def test_load_svmlight_file_descriptor(self):
        with open(datafile, 'rb') as reader:
            X, y = load_svmlight_file(reader)

        # test X's shape
        assert_array_equal(X.shape, (3, 20))

        # test y
        assert_array_equal(y, [1, 2, 3])

    def test_load_svmlight_files_comment_qid(self):
        X_train, y_train, q_train, X_test, y_test, q_test = \
            load_svmlight_files([datafile] * 2,  query_id=True)
        assert_array_equal(X_train, X_test)
        assert_array_equal(y_train, y_test)
        assert_equal(X_train.dtype, np.float32)
        assert_equal(X_test.dtype, np.float32)

        X1, y1, q1, X2, y2, q2, X3, y3, q3 = load_svmlight_files([datafile] * 3, query_id=True)
        assert_equal(X1.dtype, X2.dtype)
        assert_equal(X2.dtype, X3.dtype)
        assert_equal(X3.dtype, np.float32)

    def test_load_svmlight_files(self):
        # print load_svmlight_files([datafile] * 2)
        X_train, y_train, X_test, y_test = load_svmlight_files([datafile] * 2, query_id=False)
        assert_array_equal(X_train, X_test)
        assert_array_equal(y_train, y_test)
        assert_equal(X_train.dtype, np.float32)
        assert_equal(X_test.dtype, np.float32)

        X1, y1, X2, y2, X3, y3 = load_svmlight_files([datafile] * 3, query_id=False)
        assert_equal(X1.dtype, X2.dtype)
        assert_equal(X2.dtype, X3.dtype)
        assert_equal(X3.dtype, np.float32)

    def test_load_invalid_file(self):
        try:
            load_svmlight_file(invalidfile)
            assert False
        except RuntimeError:
            pass

    def test_load_invalid_file2(self):
        try:
            load_svmlight_files([datafile, invalidfile, datafile])
            assert False
        except RuntimeError:
            pass

    @raises(IOError)
    def test_invalid_filename(self):
        load_svmlight_file("trou pic nic douille")

    @unittest.skipIf(scikit_missing, "Scikit-Learn package missing")
    def test_dump(self):
        tmpfile = "tmp_dump.txt"
        try:
            # loads from file
            Xs, y = load_svmlight_file(datafile)

            # dumps to file
            dump_svmlight_file(Xs, y, tmpfile, zero_based=False)

            # loads them as CSR MATRIX
            X2, y2 = sk_load_svmlight_file(tmpfile)

            X3 = np.ndarray(shape=X2.shape, dtype=X2.dtype)
            X2.toarray(out=X3)

            # check assertions
            assert_array_almost_equal(Xs, X3)
            assert_array_almost_equal(y, y2)
        finally:
            if os.path.exists(tmpfile):
                os.remove(tmpfile)

    @unittest.skipIf(scikit_missing, "Scikit-Learn package missing")
    def test_dump_qid(self):
        tmpfile = "/tmp/tmp_dump.txt"
        try:
            # loads from file
            Xs, y, q = load_svmlight_file(qid_datafile, query_id=True)

            # dumps to file
            dump_svmlight_file(Xs, y, tmpfile, query_id=list(q), zero_based=False)

            # loads them as CSR MATRIX with scikit-learn
            X2, y2, q2 = sk_load_svmlight_file(tmpfile, query_id=True)

            X3 = np.ndarray(shape=X2.shape, dtype=X2.dtype)
            X2.toarray(out=X3)

            # check assertions
            assert_array_almost_equal(Xs, X3)
            assert_array_almost_equal(y, y2)
            assert_array_equal(q, q2)
        finally:
            if os.path.exists(tmpfile):
                os.remove(tmpfile)


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.DEBUG)
    unittest.main()



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Cristina Muntean <cristina.muntean@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

"""
import numpy as np
from abc import ABCMeta, abstractmethod
import six


class Metric(six.with_metaclass(ABCMeta)):
    """
    Metric is an abstract class which provides an interface for specific metrics.
    It also offers 2 methods, one for iterating over the indeces for a certain
    query and another for iterating over the entire dataset based on those
    indices.

    Some intuitions:
    https://stats.stackexchange.com/questions/159657/metrics-for-evaluating-ranking-algorithms
    """

    @abstractmethod
    def __init__(self, name):
        """
        The constructor for any metric; it initializes that metric with the
        proper name.
        
        Parameters
        ----------
        name : string
            Represents the name of that metric instance.
        """
        self.name = name
        self.detailed_scores = None

    @abstractmethod
    def eval(self, dataset, y_pred):
        """
        This abstract method computes a specific metric over the predicted
        scores for a test dataset. It calls the eval_per query method for each
        query in order to get the detailed metric score.

        Parameters
        ----------
        dataset : Dataset
            Represents the Dataset object on which we want to apply the metric.
        y_pred : numpy 1d array of float
            Represents the predicted document scores for each instance in the
            dataset.

        Returns
        -------
        avg_score: float
            Represents the average values of a metric over all metric scores
            per query.
        detailed_scores: numpy 1d array of floats
            Represents the detailed metric scores for each query. It has the
            length of n_queries.
        """
        self.detailed_scores = np.zeros(dataset.n_queries, dtype=np.float32)

        for rel_qid, (qid, q_y, q_y_pred) in enumerate(
                self.query_iterator(dataset, y_pred)):
            self.detailed_scores[rel_qid] = self.eval_per_query(q_y, q_y_pred)
        return np.nanmean(self.detailed_scores), self.detailed_scores

    @abstractmethod
    def eval_per_query(self, y, y_pred):
        """
        This methods helps to evaluate the predicted scores for a specific
        query within the dataset.

        Parameters
        ----------
        y: numpy array
            Represents the instance labels corresponding to the queries in the
            dataset (ground truth).
        y_pred: numpy array.
            Represents the predicted document scores obtained during the model
            scoring phase for that query.

        Returns
        -------
        : float
            Represents the metric score for one query.
        """

    def query_iterator(self, dataset, y_pred):
        """
        This method iterates over dataset document scores and predicted scores
        in blocks of instances which belong to the same query.
        Parameters
        ----------
        dataset :  Datatset
        y_pred  : numpy array

        Returns
        -------
        : int
            The query id.
        : numpy.array
            The document scores of the instances in the labeled dataset
            (instance labels) belonging to the same query id.
        : numpy.array
            The predicted scores for the instances in the dataset belonging to
            the same query id.
        """
        for query_id, start_offset, end_offset in dataset.query_iterator():
            yield (query_id,
                   dataset.y[start_offset:end_offset],
                   y_pred[start_offset:end_offset])



import logging
import os
import unittest

from numpy.testing import assert_equal, assert_array_equal, \
    assert_array_almost_equal

from rankeval.model import ProxyQuickRank
from rankeval.model import RTEnsemble
from rankeval.test.base import data_dir

model_file = os.path.join(data_dir, "quickrank.model.xml")


class ProxyQuickRankTestCase(unittest.TestCase):

    def setUp(self):
        self.model = RTEnsemble(model_file, format="QuickRank")

    def tearDown(self):
        del self.model
        self.model = None

    def test_count_nodes(self):
        n_trees, n_nodes = ProxyQuickRank._count_nodes(model_file)
        assert_equal(n_trees, 2)
        assert_equal(n_nodes, 10)
        assert_equal(n_trees, self.model.trees_root.size)
        assert_equal(n_nodes, self.model.trees_nodes_value.size)

    def test_root_nodes(self):
        assert_equal((self.model.trees_root > -1).all(), True,
                     "Root nodes not set correctly")

    def test_root_nodes_adv(self):
        assert_array_equal(self.model.trees_root, [0, 5],
                           "Root nodes are not correct")

    def test_tree_weights(self):
        assert_array_almost_equal(self.model.trees_weight,
                                  [0.10000000149011612, 0.10000000149011612],
                                  err_msg="Tree Weights are not correct")

    def test_split_features(self):
        assert_array_equal(self.model.trees_nodes_feature,
                           [107, 114, -1, -1, -1, 7, -1, 105, -1, -1])

    def test_tree_values(self):
        assert_array_almost_equal(self.model.trees_nodes_value,
            [14.895151138305664, -8.0245580673217773, 0.3412887828162291,
             0.66845277963831218, 0.96317280453257792, 0.66666698455810547,
             0.37133907932286642, 17.135160446166992, 0.54762687170967062,
             0.98651670670179537],
            err_msg="Split threshold values or leaf outputs are not correct")

    def test_left_children(self):
        assert_array_equal(self.model.trees_left_child,
                           [1, 2, -1, -1, -1, 6, -1, 8, -1, -1])

    def test_right_children(self):
        assert_array_equal(self.model.trees_right_child,
                           [4, 3, -1, -1, -1, 7, -1, 9, -1, -1])

    def test_leaf_correctness(self):
        for idx, feature in enumerate(self.model.trees_nodes_feature):
            if feature == -1:
                assert_equal(self.model.trees_left_child[idx], -1,
                             "Left child of a leaf node is not empty (-1)")
                assert_equal(self.model.trees_right_child[idx], -1,
                             "Right child of a leaf node is not empty (-1)")
                assert_equal(self.model.is_leaf_node(idx), True,
                             "Leaf node not detected as a leaf")

    def test_load_save_quickrank_model(self):
        # save the model
        saved_model_file = model_file + ".saved.xml"
        saved = self.model.save(saved_model_file, format="QuickRank")
        assert_equal(saved, True, "File not save correctly")

        # reload the model
        model_reloaded = RTEnsemble(saved_model_file, format="QuickRank")

        os.remove(saved_model_file)

        assert_array_almost_equal(self.model.trees_root, model_reloaded.trees_root,
                                  err_msg="Tree roots are incorrect")
        assert_array_almost_equal(self.model.trees_weight, model_reloaded.trees_weight,
                                  err_msg="Tree weights are incorrect")
        assert_array_almost_equal(self.model.trees_nodes_value, model_reloaded.trees_nodes_value,
                                  err_msg="Node thresholds are incorrect")
        assert_array_almost_equal(self.model.trees_nodes_feature, model_reloaded.trees_nodes_feature,
                                  err_msg="Node features are incorrect")
        assert_array_almost_equal(self.model.trees_left_child, model_reloaded.trees_left_child,
                                  err_msg="Left children are incorrect")
        assert_array_almost_equal(self.model.trees_right_child, model_reloaded.trees_right_child,
                                  err_msg="Right children are incorrect")


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s',
                        level=logging.DEBUG)
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 HPC lab, ISTI-CNR <salvatore.trani@isti.cnr.it>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html

"""
Run with:

sudo python ./setup.py install
"""

import sys
import os
import io
import shutil
import contextlib

from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext
from distutils.command.clean import clean as Clean
from distutils import log as _log
from distutils.dir_util import mkpath as _mkpath
import tempfile as _tempfile
from distutils.errors import CCompilerError as _CCompilerError
from distutils.version import LooseVersion
from distutils import log

min_cython_ver = '0.25.2'
try:
    import Cython
    ver = Cython.__version__
    _CYTHON_INSTALLED = ver >= LooseVersion(min_cython_ver)
except ImportError:
    _CYTHON_INSTALLED = False


@contextlib.contextmanager
def stdchannel_redirected(stdchannel, dest_filename):
    """
    A context manager to temporarily redirect stdout or stderr

    e.g.:

    with stdchannel_redirected(sys.stderr, os.devnull):
        if compiler.has_function('clock_gettime', libraries=['rt']):
            libraries.append('rt')
    """

    try:
        oldstdchannel = os.dup(stdchannel.fileno())
        dest_file = open(dest_filename, 'w')
        os.dup2(dest_file.fileno(), stdchannel.fileno())
        yield
    finally:
        if oldstdchannel is not None:
            os.dup2(oldstdchannel, stdchannel.fileno())
        if dest_file is not None:
            dest_file.close()


class custom_build_ext(build_ext):
    # the following is needed to be able to add numpy's include dirs... without
    # importing numpy directly in this script, before it's actually installed!
    # http://stackoverflow.com/questions/19919905/how-to-bootstrap-numpy-installation-in-setup-py
    def finalize_options(self):
        build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        # https://docs.python.org/2/library/__builtin__.html#module-__builtin__
        if isinstance(__builtins__, dict):
            __builtins__["__NUMPY_SETUP__"] = False
        else:
            __builtins__.__NUMPY_SETUP__ = False

        import numpy as np
        self.include_dirs.append(np.get_include())

    def _check_openmp_support(self):
        # Compile a test program to determine if compiler supports OpenMP.
        _mkpath(self.build_temp)
        with stdchannel_redirected(sys.stderr, os.devnull):
            with _tempfile.NamedTemporaryFile(mode='w',
                                              dir=self.build_temp,
                                              prefix='openmptest',
                                              suffix='.c') as srcfile:
                _log.info("checking if compiler supports OpenMP")
                srcfile.write("""
                #include <omp.h>
                int testfunc() {
                    int i;
                    #pragma omp parallel for
                    for (i = 0; i < 10; i ++)
                        ;
                    return omp_get_num_threads();
                }
                """)
                srcfile.flush()
                try:
                    objects = self.compiler.compile([srcfile.name],
                                                    extra_postargs=["-fopenmp"],
                                                    output_dir="/")
                except _CCompilerError:
                    _log.info("compiler does not support OpenMP")
                    use_openmp = False
                else:
                    _log.info("enabling OpenMP support")
                    use_openmp = True
                    for o in objects:
                        os.remove(o)
            return use_openmp

    def build_extensions(self):
        use_openmp = self._check_openmp_support()
        if use_openmp:
            for ext in self.extensions:
                if not ext.extra_compile_args:
                    ext.extra_compile_args = []
                ext.extra_compile_args.append('-fopenmp')
                if not ext.extra_link_args:
                    ext.extra_link_args = []
                ext.extra_link_args.append('-fopenmp')

        # Add language level directive (for the cython compiler)
        for ext in self.extensions:
            ext.compiler_directives = {'language_level' : sys.version_info[0]}

        # Chain to method in parent class.
        build_ext.build_extensions(self)
        if not _CYTHON_INSTALLED:
            log.info('No supported version of Cython installed. '
                     'Installing from compiled files.')
            for extension in self.extensions:
                sources = []
                for sfile in extension.sources:
                    path, ext = os.path.splitext(sfile)
                    if ext in ('.pyx', '.py'):
                        if extension.language == 'c++':
                            ext = '.cpp'
                        else:
                            ext = '.c'
                        sfile = path + ext
                    sources.append(sfile)
                extension.sources[:] = sources


# Custom clean command to remove build artifacts
class CleanCommand(Clean):
    description = "Remove build artifacts from the source tree"

    def run(self):
        Clean.run(self)
        # Remove c files if we are not within a sdist package
        cwd = os.path.abspath(os.path.dirname(__file__))
        remove_c_files = not os.path.exists(os.path.join(cwd, 'PKG-INFO'))
        if remove_c_files:
            print('Will remove generated .c files')
        if os.path.exists('build'):
            shutil.rmtree('build')
        for dirpath, dirnames, filenames in os.walk('rankeval'):
            for filename in filenames:
                if any(filename.endswith(suffix) for suffix in
                       (".so", ".pyd", ".dll", ".pyc")):
                    os.unlink(os.path.join(dirpath, filename))
                    continue
                extension = os.path.splitext(filename)[1]
                if remove_c_files and extension in ['.c', '.cpp']:
                    pyx_file = str.replace(filename, extension, '.pyx')
                    if os.path.exists(os.path.join(dirpath, pyx_file)):
                        os.unlink(os.path.join(dirpath, filename))
            for dirname in dirnames:
                if dirname == '__pycache__':
                    shutil.rmtree(os.path.join(dirpath, dirname))


root_dir = os.path.dirname(os.path.abspath(os.path.expanduser(__file__)))
# make it relative in order to avoid errors of absolute path by pypi
root_dir = os.path.relpath(root_dir)
rankeval_dir = os.path.join(root_dir, 'rankeval')
dataset_dir = os.path.join(rankeval_dir, 'dataset')
scoring_dir = os.path.join(rankeval_dir, 'scoring')
analysis_dir = os.path.join(rankeval_dir, 'analysis')

cmdclass = {
    'build_ext': custom_build_ext,
    'clean': CleanCommand
}

version = io.open(os.path.join(rankeval_dir, 'VERSION'),
                  encoding='utf-8').read().strip()
readme = io.open(os.path.join(root_dir, 'README.md'), encoding='utf-8').read()

setup(
    name='rankeval',
    version=version,
    description='Tool for the analysis and evaluation of Learning to Rank '
                'models based on ensembles of regression trees.',
    long_description=readme,
    long_description_content_type='text/markdown',

    ext_modules=[
        Extension('rankeval.dataset._svmlight_format',
                  sources=[dataset_dir + '/_svmlight_format.cpp'],
                  include_dirs=[dataset_dir],
                  language='c++',
                  extra_compile_args=['-O3', "-std=c++11"]),
        Extension('rankeval.scoring._efficient_scoring',
                  sources=[scoring_dir + '/_efficient_scoring.pyx'],
                  include_dirs=[scoring_dir],
                  extra_compile_args=['-O3'],
                  ),
        Extension('rankeval.analysis._efficient_topological',
                  sources=[analysis_dir + '/_efficient_topological.pyx'],
                  include_dirs=[analysis_dir],
                  extra_compile_args=['-O3'],
                  ),
        Extension('rankeval.analysis._efficient_feature',
                  sources=[analysis_dir + '/_efficient_feature.pyx',
                           analysis_dir + '/_efficient_feature_impl.cpp'],
                  include_dirs=[analysis_dir],
                  language='c++',
                  extra_compile_args=['-O3', "-w", "-std=c++11"],
                  )
    ],

    cmdclass=cmdclass,
    license='MPL 2.0',
    packages=find_packages(),

    author="HPC lab, ISTI-CNR",
    author_email='rankeval@isti.cnr.it',

    url='http://rankeval.isti.cnr.it',
    download_url='http://pypi.python.org/pypi/rankeval',

    keywords='Learning to Rank, Model Analysis, Model Evaluation, '
             'Ensemble of Regression Trees',

    platforms='any',

    zip_safe=False,

    classifiers=[  # from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Framework :: Jupyter',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Benchmark'
    ],

    # py_modules=['svmlight_loader',],

    test_suite="rankeval.test",
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    setup_requires=[
        'Cython >= {}'.format(min_cython_ver),
        'setuptools >= 24.2.0',
        'numpy >= 1.13',
        'scipy >= 0.7.0'
    ],
    install_requires=[
        # Use 1.13: https://github.com/quantopian/zipline/issues/1808
        'numpy >= 1.13',
        'scipy >= 0.14.0',
        'six >= 1.9.0',
        'pandas >= 0.19.1',
        'xarray >= 0.10.9',
        'seaborn >= 0.8',
        'matplotlib >= 2.0.2',
    ],
    tests_require=[
        'nose >= 1.3.7',
    ],

    extras_require={
        'develop': [
            'sphinx >= 1.5.0',
            'sphinx_rtd_theme >= 0.2.0',
            'numpydoc > 0.5.0',
        ],
    },

    include_package_data=True,
)



# Copyright (c) 2017, All Contributors (see CONTRIBUTORS file)
# Authors: Salvatore Trani <salvatore.trani@isti.cnr.it>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
The :mod:`rankeval.model` module includes utilities to load a model
and dump it according to several supported model's format.
"""

from .proxy_LightGBM import ProxyLightGBM
from .proxy_QuickRank import ProxyQuickRank
from .proxy_ScikitLearn import ProxyScikitLearn
from .proxy_XGBoost import ProxyXGBoost
from .proxy_Jforests import ProxyJforests
from .proxy_CatBoost import ProxyCatBoost
from .rt_ensemble import RTEnsemble

__all__ = ['RTEnsemble',
           'ProxyQuickRank',
           'ProxyLightGBM',
           'ProxyXGBoost',
           'ProxyScikitLearn',
           'ProxyJforests',
           'ProxyCatBoost'
]
